# This file is part of MyPaint.
# Copyright (C) 2010-2018 by the MyPaint Development Team
# Copyright (C) 2009-2013 by Martin Renold <martinxyz@gmx.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

## Imports

from __future__ import division, print_function
from gettext import gettext as _
import colorsys

import gui.mode
from .overlays import Overlay
from .overlays import rounded_box
import lib.color


## Color picking mode, with a preview rectangle overlay

class ColorPickMode (gui.mode.OneshotDragMode):
    """Mode for picking colors from the screen, with a preview

    This can be invoked in quite a number of ways:

    * The keyboard hotkey ("R" by default)
    * Modifier and pointer button: (Ctrl+Button1 by default)
    * From the toolbar or menu

    The first two methods pick immediately. Moving the mouse with the
    initial keys or buttons held down keeps picking with a little
    preview square appearing.

    The third method doesn't pick immediately: you have to click on the
    canvas to start picking.

    While the preview square is visible, it's possible to pick outside
    the window. This "hidden" functionality may not work at all with
    more modern window managers and DEs, and may be removed if it proves
    slow or faulty.

    """
    # Class configuration
    ACTION_NAME = 'ColorPickMode'
    PICK_SIZE = 6

    # Keyboard activation behaviour (instance defaults)
    # See keyboard.py and doc.mode_flip_action_activated_cb()
    keyup_timeout = 0   # don't change behaviour by timeout

    pointer_behavior = gui.mode.Behavior.EDIT_OBJECTS
    scroll_behavior = gui.mode.Behavior.NONE
    # XXX ^^^^^^^ grabs ptr, so no CHANGE_VIEW
    supports_button_switching = False

    @property
    def inactive_cursor(self):
        return self.doc.app.cursor_color_picker

    @classmethod
    def get_name(cls):
        return _(u"Pick Color")

    def get_usage(self):
        return _(u"Set the color used for painting")

    def __init__(self, ignore_modifiers=False, pickmode="PickAll", **kwds):
        super(ColorPickMode, self).__init__(**kwds)
        self._overlay = None
        self._started_from_key_press = ignore_modifiers
        self._start_drag_on_next_motion_event = False
        self._pickmode = pickmode

    def enter(self, doc, **kwds):
        """Enters the mode, arranging for necessary grabs ASAP"""
        super(ColorPickMode, self).enter(doc, **kwds)
        if self._started_from_key_press:
            # Pick now using the last recorded event position
            doc = self.doc
            tdw = self.doc.tdw
            t, x, y = doc.get_last_event_info(tdw)
            if None not in (x, y):
                self._pick_color_mode(tdw, x, y, self._pickmode)
            # Start the drag when possible
            self._start_drag_on_next_motion_event = True
            self._needs_drag_start = True

    def leave(self, **kwds):
        self._remove_overlay()
        super(ColorPickMode, self).leave(**kwds)

    def button_press_cb(self, tdw, event):
        self._pick_color_mode(tdw, event.x, event.y, self._pickmode)
        # Supercall will start the drag normally
        self._start_drag_on_next_motion_event = False
        return super(ColorPickMode, self).button_press_cb(tdw, event)

    def motion_notify_cb(self, tdw, event):
        if self._start_drag_on_next_motion_event:
            self._start_drag(tdw, event)
            self._start_drag_on_next_motion_event = False
        return super(ColorPickMode, self).motion_notify_cb(tdw, event)

    def drag_stop_cb(self, tdw):
        self._remove_overlay()
        super(ColorPickMode, self).drag_stop_cb(tdw)

    def drag_update_cb(self, tdw, event, dx, dy):
        self._pick_color_mode(tdw, event.x, event.y, self._pickmode)
        self._place_overlay(tdw, event.x, event.y)
        return super(ColorPickMode, self).drag_update_cb(tdw, event, dx, dy)

    def _place_overlay(self, tdw, x, y):
        if self._overlay is None:
            self._overlay = ColorPickPreviewOverlay(self.doc, tdw, x, y)
        else:
            self._overlay.move(x, y)

    def _remove_overlay(self):
        if self._overlay is None:
            return
        self._overlay.cleanup()
        self._overlay = None

    def get_options_widget(self):
        return None

    def _pick_color_mode(self, tdw, x, y, mode):
        # Init shared variables between normal and HCY modes.
        pickcolor = tdw.pick_color(x, y)
        brushcolor = self._get_app_brush_color()
        brushcolor_rgb = brushcolor.get_rgb()
        pickcolor_rgb = pickcolor.get_rgb()

        # If brush and pick colors are the same, nothing to do.
        if brushcolor_rgb != pickcolor_rgb:
            pickcolor_hsv = pickcolor.get_hsv()
            brushcolor_hsv = brushcolor.get_hsv()
            cm = self.doc.app.brush_color_manager
            c_min = 0.0001
            y_min = 0.0001
            y_max = 0.9999

            # Normal pick mode, but preserve hue for achromatic colors.
            # Easy because we are staying in HSV
            # and not converting to another color space.

            if mode == "PickAll":
                if (pickcolor_hsv[1] == 0) or (pickcolor_hsv[2] == 0):
                    brushcolornew_hsv = lib.color.HSVColor(
                        brushcolor_hsv[0],
                        pickcolor_hsv[1],
                        pickcolor_hsv[2],
                    )
                    pickcolor = brushcolornew_hsv
                cm.set_color(pickcolor)
            else:
                # Pick H, C, or Y independently.
                # Deal with scenarios to avoid achromatic conversions
                # from HCY to RGB to HSV losing hue information.
                # Basically prevent C from reaching 0
                # and Y from reaching 0.0 or 1.0.
                brushcolor_hcy = lib.color.RGB_to_HCY(brushcolor_rgb)
                pickcolor_hcy = lib.color.RGB_to_HCY(pickcolor_rgb)

                if mode == "PickHue":
                    brushcolornew_hcy = (
                        pickcolor_hcy[0],
                        brushcolor_hcy[1],
                        brushcolor_hcy[2],
                    )
                    if brushcolor_hcy[1] < c_min:
                        brushcolornew_hcy = (
                            brushcolornew_hcy[0],
                            c_min,
                            brushcolornew_hcy[2],
                        )
                    if pickcolor_hcy[2] < y_min:
                        brushcolornew_hcy = (
                            brushcolornew_hcy[0],
                            brushcolornew_hcy[1],
                            y_min,
                        )
                    if pickcolor_hcy[2] > y_max:
                        brushcolornew_hcy = (
                            brushcolornew_hcy[0],
                            brushcolornew_hcy[1],
                            y_max,
                        )
                    if (pickcolor_hsv[1] == 0) or (pickcolor_hsv[2] == 0):
                        brushcolornew_hcy = (
                            brushcolor_hsv[0],
                            brushcolor_hcy[1],
                            brushcolor_hcy[2],
                        )
                elif mode == "PickLuma":
                    brushcolornew_hcy = (
                        brushcolor_hsv[0],
                        brushcolor_hcy[1],
                        pickcolor_hcy[2],
                    )
                    if brushcolor_hcy[1] < c_min:
                        brushcolornew_hcy = (
                            brushcolornew_hcy[0],
                            c_min,
                            pickcolor_hcy[2],
                        )
                    if pickcolor_hcy[2] < y_min:
                        brushcolornew_hcy = (
                            brushcolornew_hcy[0],
                            brushcolornew_hcy[1],
                            y_min,
                        )
                    if pickcolor_hcy[2] > y_max:
                        brushcolornew_hcy = (
                            brushcolornew_hcy[0],
                            brushcolornew_hcy[1],
                            y_max,
                        )
                elif mode == "PickChroma":
                    brushcolornew_hcy = (
                        brushcolor_hsv[0],
                        pickcolor_hcy[1],
                        brushcolor_hcy[2],
                    )
                    if pickcolor_hcy[1] < c_min:
                        brushcolornew_hcy = (
                            brushcolornew_hcy[0],
                            c_min,
                            brushcolornew_hcy[2],
                        )
                    if brushcolor_hcy[2] < y_min:
                        brushcolornew_hcy = (
                            brushcolornew_hcy[0],
                            brushcolornew_hcy[1],
                            y_min,
                        )
                    if brushcolor_hcy[2] > y_max:
                        brushcolornew_hcy = (
                            brushcolornew_hcy[0],
                            brushcolornew_hcy[1],
                            y_max,
                        )

                brushcolornew = lib.color.HCY_to_RGB(brushcolornew_hcy)
                brushcolornew = colorsys.rgb_to_hsv(*brushcolornew)
                brushcolornew = lib.color.HSVColor(*brushcolornew)
                cm.set_color(brushcolornew)
        return None

    def _get_app_brush_color(self):
        app = self.doc.app
        return lib.color.HSVColor(*app.brush.get_color_hsv())


class ColorPickModeH (ColorPickMode):

    # Class configuration
    ACTION_NAME = 'ColorPickModeH'

    @property
    def inactive_cursor(self):
        return self.doc.app.cursor_color_picker_h

    @classmethod
    def get_name(cls):
        return _(u"Pick Hue")

    def get_usage(self):
        return _(u"Set the color Hue used for painting")

    def __init__(self, ignore_modifiers=False, pickmode="PickHue", **kwds):
        super(ColorPickModeH, self).__init__(**kwds)
        self._overlay = None
        self._started_from_key_press = ignore_modifiers
        self._start_drag_on_next_motion_event = False
        self._pickmode = pickmode


class ColorPickModeC (ColorPickMode):
    # Class configuration
    ACTION_NAME = 'ColorPickModeC'

    @property
    def inactive_cursor(self):
        return self.doc.app.cursor_color_picker_c

    @classmethod
    def get_name(cls):
        return _(u"Pick Chroma")

    def get_usage(self):
        return _(u"Set the color Chroma used for painting")

    def __init__(self, ignore_modifiers=False, pickmode="PickChroma", **kwds):
        super(ColorPickModeC, self).__init__(**kwds)
        self._overlay = None
        self._started_from_key_press = ignore_modifiers
        self._start_drag_on_next_motion_event = False
        self._pickmode = pickmode


class ColorPickModeY (ColorPickMode):
    # Class configuration
    ACTION_NAME = 'ColorPickModeY'

    @property
    def inactive_cursor(self):
        return self.doc.app.cursor_color_picker_y

    @classmethod
    def get_name(cls):
        return _(u"Pick Luma")

    def get_usage(self):
        return _(u"Set the color Luma used for painting")

    def __init__(self, ignore_modifiers=False, pickmode="PickLuma", **kwds):
        super(ColorPickModeY, self).__init__(**kwds)
        self._overlay = None
        self._started_from_key_press = ignore_modifiers
        self._start_drag_on_next_motion_event = False
        self._pickmode = pickmode


class ColorPickPreviewOverlay (Overlay):
    """Preview overlay during color picker mode.

    This is only shown when dragging the pointer with a button or the
    hotkey held down, to avoid flashing and distraction.

    """

    PREVIEW_SIZE = 70
    OUTLINE_WIDTH = 3
    CORNER_RADIUS = 10

    def __init__(self, doc, tdw, x, y):
        """Initialize, attaching to the brush and to the tdw.

        Observer callbacks and canvas overlays are registered by this
        constructor, so cleanup() must be called when the owning mode leave()s.

        """
        Overlay.__init__(self)
        self._doc = doc
        self._tdw = tdw
        self._x = int(x)+0.5
        self._y = int(y)+0.5
        alloc = tdw.get_allocation()
        self._tdw_w = alloc.width
        self._tdw_h = alloc.height
        self._color = self._get_app_brush_color()
        app = doc.app
        app.brush.observers.append(self._brush_color_changed_cb)
        tdw.display_overlays.append(self)
        self._previous_area = None
        self._queue_tdw_redraw()

    def cleanup(self):
        """Cleans up temporary observer stuff, allowing garbage collection.
        """
        app = self._doc.app
        app.brush.observers.remove(self._brush_color_changed_cb)
        self._tdw.display_overlays.remove(self)
        assert self._brush_color_changed_cb not in app.brush.observers
        assert self not in self._tdw.display_overlays
        self._queue_tdw_redraw()

    def move(self, x, y):
        """Moves the preview square to a new location, in tdw pointer coords.
        """
        self._x = int(x)+0.5
        self._y = int(y)+0.5
        self._queue_tdw_redraw()

    def _get_app_brush_color(self):
        app = self._doc.app
        return lib.color.HSVColor(*app.brush.get_color_hsv())

    def _brush_color_changed_cb(self, settings):
        if not settings.intersection(('color_h', 'color_s', 'color_v')):
            return
        self._color = self._get_app_brush_color()
        self._queue_tdw_redraw()

    def _queue_tdw_redraw(self):
        if self._previous_area is not None:
            self._tdw.queue_draw_area(*self._previous_area)
            self._previous_area = None
        area = self._get_area()
        if area is not None:
            self._tdw.queue_draw_area(*area)

    def _get_area(self):
        # Returns the drawing area for the square
        size = self.PREVIEW_SIZE

        # Start with the pointer location
        x = self._x
        y = self._y

        offset = size // 2

        # Only show if the pointer is inside the tdw
        alloc = self._tdw.get_allocation()
        if x < 0 or y < 0 or y > alloc.height or x > alloc.width:
            return None

        # Convert to preview location
        # Pick a direction - N,W,E,S - in which to offset the preview
        if y + size > alloc.height - offset:
            x -= offset
            y -= size + offset
        elif x < offset:
            x += offset
            y -= offset
        elif x > alloc.width - offset:
            x -= size + offset
            y -= offset
        else:
            x -= offset
            y += offset

        ## Correct to place within the tdw
        #   if x < 0:
        #       x = 0
        #   if y < 0:
        #       y = 0
        #   if x + size > alloc.width:
        #       x = alloc.width - size
        #   if y + size > alloc.height:
        #       y = alloc.height - size

        return (int(x), int(y), size, size)

    def paint(self, cr):
        area = self._get_area()
        if area is not None:
            x, y, w, h = area

            cr.set_source_rgb(*self._color.get_rgb())
            x += (self.OUTLINE_WIDTH // 2) + 1.5
            y += (self.OUTLINE_WIDTH // 2) + 1.5
            w -= self.OUTLINE_WIDTH + 3
            h -= self.OUTLINE_WIDTH + 3
            rounded_box(cr, x, y, w, h, self.CORNER_RADIUS)
            cr.fill_preserve()

            cr.set_source_rgb(0, 0, 0)
            cr.set_line_width(self.OUTLINE_WIDTH)
            cr.stroke()

        self._previous_area = area



# This file is part of MyPaint.
# Copyright (C) 2012-2018 by the MyPaint Development Team.
# Copyright (C) 2007-2014 by Martin Renold <martinxyz@gmx.ch>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import division, print_function

import numpy as np

from . import brush


class Stroke (object):
    """Replayable record of a stroke's data

    Stroke recording objects store all information required to replay a stroke
    with the brush engine, event by event. This includes the RNG seed etc.

    A "finished" stroke object is immutable, except right after creation (when
    it has never been fully rendered).  To modify an existing stroke, the old
    one must be deleted and a new Stroke instance must be used to replace it.
    """

    _SERIAL_NUMBER = 0

    def __init__(self):
        """Initialize"""
        super(Stroke, self).__init__()
        self.finished = False
        self._SERIAL_NUMBER += 1
        self.serial_number = self._SERIAL_NUMBER

    def start_recording(self, brush):
        assert not self.finished

        bi = brush.brushinfo
        self.brush_settings = bi.save_to_string()
        self.brush_name = bi.get_string_property("parent_brush_name")

        states = brush.get_states_as_array()
        assert states.dtype == 'float32'
        self.brush_state = states.tostring()

        self.brush = brush
        self.brush.new_stroke()  # resets the stroke_* members of the brush

        self.tmp_event_list = []

    def record_event(self, dtime, x, y, pressure, xtilt, ytilt,
                     viewzoom, viewrotation, barrel_rotation):
        assert not self.finished
        self.tmp_event_list.append((dtime, x, y, pressure, xtilt, ytilt,
                                    viewzoom, viewrotation, barrel_rotation))

    def stop_recording(self):
        if self.finished:
            return
        # OPTIMIZE
        # - for space: just gzip? use integer datatypes?
        # - for time: maybe already use array storage while recording?
        data = np.array(self.tmp_event_list, dtype='float64')
        data = data.tostring()
        version = b'2'
        self.stroke_data = version + data

        self.total_painting_time = self.brush.get_total_stroke_painting_time()
        del self.brush, self.tmp_event_list
        self.finished = True

    def is_empty(self):
        return self.total_painting_time == 0

    empty = property(is_empty)

    def render(self, surface):
        assert self.finished

        # OPTIMIZE: check if parsing of settings is a performance bottleneck
        b = brush.Brush(brush.BrushInfo(self.brush_settings))

        states = np.fromstring(self.brush_state, dtype='float32')
        b.set_states_from_array(states)

        version, data = self.stroke_data[0:1], self.stroke_data[1:]
        assert version == b'2'
        data = np.fromstring(data, dtype='float64')
        data.shape = (len(data) // 9, 9)

        surface.begin_atomic()
        for (dtime, x, y, pressure, xtilt, ytilt, viewzoom,
             viewrotation, barrel_rotation) in data:
            b.stroke_to(surface.backend, x, y, pressure, xtilt, ytilt, dtime,
                        viewzoom, viewrotation, barrel_rotation)
        surface.end_atomic()

    def copy_using_different_brush(self, brushinfo):
        assert self.finished
        # Make a shallow clone of almost everything
        clone = Stroke()
        clone.__dict__.update(self.__dict__)
        # Except for the brush-specific stuff
        clone.brush_settings = brushinfo.save_to_string()
        clone.brush_name = brushinfo.get_string_property("parent_brush_name")
        # note: we keep self.brush_state intact, even if the new brush
        # has different meanings for the states. This should cause
        # fewer glitches than resetting the initial state to zero.
        return clone



# This file is part of MyPaint.
# Copyright (C) 2014-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or

"""Canvas input modes API: stack, and base classes for modes."""


## Imports

from __future__ import division, print_function
import logging
import math
from gettext import gettext as _

from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import GLib

import lib.command
from lib.document import Document
from lib.layer.data import SimplePaintingLayer
from lib.observable import event
from lib.pycompat import add_metaclass
from lib.pycompat import unicode

import gui.cursor

logger = logging.getLogger(__name__)

## Module constants

# Actions it makes sense to bind to a button. (HACK).
# Notably, tablet pads tend to offer many more buttons than the usual 3...

BUTTON_BINDING_ACTIONS = [
    "ShowPopupMenu",
    "Undo",
    "Redo",
    "Bigger",
    "Smaller",
    "MoreOpaque",
    "LessOpaque",
    "PickContext",
    "Fullscreen",
    "ToggleSubwindows",
    "BrushChooserPopup",
    "ColorChooserPopup",
    "ColorDetailsDialog",
    "ColorHistoryPopup",
    "PalettePrev",
    "PaletteNext",
]


## Behaviour flags

class Behavior:
    """Broad classification of what a mode's handler methods do

    These flags are assigned to devices in `gui.device` to allow the
    user to limit what devices are allowed to do. Mode instances expose
    their behaviour by defining their pointer_behavior and
    scroll_behavior properties.

    """
    NONE = 0x00  #: the mode does not perform any action
    PAINT_FREEHAND = 0x01  #: paint freehand brushstrokes
    PAINT_LINE = 0x02  #: non-freehand brushstrokes: lines, perspective
    EDIT_OBJECTS = 0x04  #: move and adjust objects on screen
    CHANGE_VIEW = 0x08  #: move the viewport around
    PAINT_NOBRUSH = 0x10   #: painting independent of brush (eg. fill)
    # Useful masks
    PAINT_BRUSH = PAINT_FREEHAND | PAINT_LINE  #: painting dependent of brush
    PAINT_CONSTRAINED = PAINT_LINE | PAINT_NOBRUSH  #: non-freehand painting
    NON_PAINTING = EDIT_OBJECTS | CHANGE_VIEW
    ALL_PAINTING = PAINT_FREEHAND | PAINT_CONSTRAINED
    ALL = NON_PAINTING | ALL_PAINTING


## Metaclass for all modes

class ModeRegistry (type):
    """Lookup table for interaction modes and their associated actions

    Operates as the metaclass for `InteractionMode`, so all you need to do to
    create the association for a mode subclass is to define an
    ``ACTION_NAME`` entry in the class's namespace containing the name of
    the associated `Gtk.Action` defined in ``resources.xml``.

    """

    action_name_to_mode_class = {}
    mode_classes = set()

    # (Special-cased @staticmethod)
    def __new__(cls, name, bases, dict):
        """Creates and records a new (InteractionMode) class.

        :param cls: this metaclass
        :param name: name of the class under construction
        :param bases: immediate base classes of the class under construction
        :param dict: class dict for the class under construction
        :rtype: the constructed class, a regular InteractionMode class object

        If it exists, the ``ACTION_NAME`` entry in `dict` is recorded,
        and can be used as a key for lookup of the returned class via the
        ``@classmethod``s defined on `ModeRegistry`.

        """
        action_name = dict.get("ACTION_NAME", None)
        mode_class = super(ModeRegistry, cls).__new__(cls, name, bases, dict)
        if action_name is not None:
            action_name = str(action_name)
            cls.action_name_to_mode_class[action_name] = mode_class
        cls.mode_classes.add(mode_class)
        return mode_class

    @classmethod
    def get_mode_class(cls, action_name):
        """Looks up a registered mode class by its associated action's name.

        :param action_name: a string containing an action name (see this
           metaclass's docs regarding the ``ACTION_NAME`` class variable)
        :rtype: an InteractionMode class object, or `None`.

        """
        return cls.action_name_to_mode_class.get(action_name, None)

    @classmethod
    def get_action_names(cls):
        """Returns all action names associated with interaction.

        :rtype: an iterable of action name strings.

        """
        return cls.action_name_to_mode_class.keys()


## Mode base classes

@add_metaclass(ModeRegistry)
class InteractionMode (object):
    """Required base class for temporary interaction modes.

    Active interaction mode objects process input events, and can manipulate
    document views (TiledDrawWidget), the document model data (lib.document),
    and the mode stack they sit on. Interactions encapsulate state about their
    particular kind of interaction; for example a drag interaction typically
    contains the starting position for the drag.

    Event handler methods can create new sub-modes and push them to the stack.
    It is conventional to pass the current event to the equivalent method on
    the new object when this transfer of control happens.

    Subclasses may nominate a related `GtkAction` instance in the UI by setting
    the class-level variable ``ACTION_NAME``: this should be the name of an
    action defined in `gui.app.Application.builder`'s XML file.

    All InteractionMode subclasses register themselves in ModeRegistry
    when their class is defined.

    """

    ## Class configuration

    #: See the docs for `gui.mode.ModeRegistry`.
    ACTION_NAME = None

    #: True if the mode supports live update from the brush editor
    IS_LIVE_UPDATEABLE = False

    #: Timeout for Document.mode_flip_action_activated_cb(). How long, in
    #: milliseconds, it takes for the controller to change the key-up action
    #: when activated with a keyboard "Flip<ModeName>" action. Set to zero
    #: for modes where key-up should exit the mode at any time, and to a larger
    #: number for modes where the behaviour changes.
    keyup_timeout = 500

    ## Defaults for instances (sue me, I'm lazy)

    #: The `gui.document.Document` this mode affects: see enter()
    doc = None

    #: Broad description of what result clicking, moving the pointer,
    #: or dragging has in this mode. See `Behavior`.
    pointer_behavior = Behavior.NONE

    #: Broad description of what result scrolling a mouse scroll-wheel
    #: does in this mode. See `Behavior`.
    scroll_behavior = Behavior.NONE

    #: True if the mode supports switching to another mode based on
    #: combinations of pointer buttons and modifier keys.
    supports_button_switching = True

    #: Optional whitelist of the names of the modes which this mode can
    #: switch to. If the iterable is empty, all modes are possible.
    permitted_switch_actions = ()

    ## Status message info

    @classmethod
    def get_name(cls):
        """Returns a short human-readable description of the mode.

        :rtype: unicode

        This is used for status bar messages, and potentially elsewhere before
        the mode has been instantiated.  All concrete subclasses should
        override this.  By default the (non-localized) class name is returned.

        When capitalizing, use whatever style the GNOME HIG specifies for menu
        items.  In English, this is currently "header", or title case (first
        word capitalized, all other words capitalized except closed-class
        words).  Do not use trailing punctuation.

        """
        return unicode(cls.__name__)

    def get_usage(self):
        """Returns a medium-length usage message for the mode.

        :rtype: unicode

        This is used for status bar messages.  All concrete subclasses should
        override this.  The default return value is an empty string.

        The usage message should be a short, natural-sounding explanation to
        the user detailing what the current mode is for.  Note that the usage
        message is typically displayed after the mode's name or explanatory
        icon, so there is no need to repeat that.  Brevity is important because
        space is limited.

        When capitalizing, use whatever style the GNOME HIG specifies for
        tooltips.  In English, this is currently sentence case.  Use one
        complete sentence, and always omit the the trailing period.

        """
        return u""

    def __unicode__(self):
        return self.get_name()

    ## Associated action

    def get_action(self):
        """Returns any app action associated with the mode."""
        if self.doc and hasattr(self.doc, "app"):
            if self.ACTION_NAME:
                return self.doc.app.find_action(self.ACTION_NAME)

    ## Mode icon

    def get_icon_name(self):
        """Returns the icon to use when representing the mode.

        If there's an associated action, this method returns the icon
        associated with the action.
        """
        icon_name = None
        action = self.get_action()
        if action:
            icon_name = action.get_icon_name()
        if not icon_name:
            return 'missing-icon'
        return icon_name

    ## Mode stacking interface

    def stackable_on(self, mode):
        """Tests whether the mode can usefully stack onto an active mode.

        :param mode: another mode object
        :rtype: bool

        This method should return True if this mode can usefully be stacked
        onto another mode when switching via toolbars buttons or other actions.

        """
        # By default, anything can be stacked on brush tools, except for brush
        # tools.
        # Why? So whenever the user picks a brush (from a shortcut or
        # whatever), the active tool becomes the last used brush-sensitive
        # tool.
        # See issue #530.
        if self.pointer_behavior & Behavior.PAINT_BRUSH:
            return False
        return mode.pointer_behavior & Behavior.PAINT_BRUSH

    def enter(self, doc, **kwds):
        """Enters the mode: called by `ModeStack.push()` etc.

        :param doc: the `gui.document.Document` this mode should affect.
            A reference is kept in `self.doc`.

        This is called when the mode becomes active, i.e. when it becomes the
        top mode on a ModeStack, and before input is sent to it. Note that a
        mode may be entered only to be left immediately: mode stacks are
        cleared by repeated pop()ing.

        """
        self.doc = doc
        assert not hasattr(super(InteractionMode, self), "enter")

    def leave(self):
        """Leaves the mode: called by `ModeStack.pop()` etc.

        This is called when an active mode becomes inactive, i.e. when it is
        no longer the top mode on its ModeStack. It should commit any
        uncommitted work to the undo stack, just as `checkpoint()` does.

        """
        self.doc = None
        assert not hasattr(super(InteractionMode, self), "leave")

    def checkpoint(self, **kwargs):
        """Commits any of the mode's uncommitted work

        This is called on the active mode at various times to signal
        that pending work should be committed to the command stack now,
        so that the Undo command would be able to undo it if it were
        called next.

        The mode continues to be active.

        This method is not automatically invoked when changing modes:
        leave() should manage that transition.

        """
        assert not hasattr(super(InteractionMode, self), "checkpoint")

    ## Event handler defaults (no-ops)

    def button_press_cb(self, tdw, event):
        """Handler for ``button-press-event``s."""
        assert not hasattr(super(InteractionMode, self), "button_press_cb")

    def motion_notify_cb(self, tdw, event):
        """Handler for ``motion-notify-event``s."""
        assert not hasattr(super(InteractionMode, self), "motion_notify_cb")

    def button_release_cb(self, tdw, event):
        """Handler for ``button-release-event``s."""
        assert not hasattr(super(InteractionMode, self), "button_release_cb")

    def scroll_cb(self, tdw, event):
        """Handler for ``scroll-event``s.
        """
        assert not hasattr(super(InteractionMode, self), "scroll_cb")

    def key_press_cb(self, win, tdw, event):
        """Handler for ``key-press-event``s.

        The base class implementation does nothing.
        Keypresses are received by the main window only, but at this point it
        has applied some heuristics to determine the active doc and view.
        These are passed through to the active mode and are accessible to
        keypress handlers via `self.doc` and the `tdw` argument.

        """
        assert not hasattr(super(InteractionMode, self), "key_press_cb")
        return True

    def key_release_cb(self, win, tdw, event):
        """Handler for ``key-release-event``s.

        The base class implementation does nothing. See `key_press_cb` for
        details of the additional arguments.

        """
        assert not hasattr(super(InteractionMode, self), "key_release_cb")
        return True

    ## Drag sub-API (FIXME: this is in the wrong place)
    # Defined here to allow mixins to provide behaviour for both both drags and
    # regular events without having to derive from DragMode. Really these
    # buck-stops-here definitions belong in DragMode, so consider moving them
    # somewhere more sensible.

    def drag_start_cb(self, tdw, event):
        assert not hasattr(super(InteractionMode, self), "drag_start_cb")

    def drag_update_cb(self, tdw, event, dx, dy):
        assert not hasattr(super(InteractionMode, self), "drag_update_cb")

    def drag_stop_cb(self, tdw):
        assert not hasattr(super(InteractionMode, self), "drag_stop_cb")

    ## Internal utility functions

    def current_modifiers(self):
        """Returns the current set of modifier keys as a Gdk bitmask.

        See: gui.document.Document.get_current_modifiers()

        """
        doc = self.doc
        if doc is None:
            modifiers = Gdk.ModifierType(0)
        else:
            modifiers = doc.get_current_modifiers()
        return modifiers

    def current_position(self):
        """Returns the current client pointer position on the main TDW.

        For use in enter() methods: since the mode may be being entered
        by the user pressing a key, no position is available at this
        point. Normal event handlers should use their argument GdkEvents
        to determine position.

        """
        disp = self.doc.tdw.get_display()
        mgr = disp.get_device_manager()
        dev = mgr.get_client_pointer()
        win = self.doc.tdw.get_window()
        underwin, x, y, mods = win.get_device_position(dev)
        return x, y


class ScrollableModeMixin (InteractionMode):
    """Mixin for scrollable modes.

    Implements some immediate rotation and zoom commands for the scroll wheel.
    These should be useful in many modes, but perhaps not all.

    """

    # Hack conversion factor from smooth scroll units to screen pixels.
    _PIXELS_PER_SMOOTH_SCROLL_UNIT = 25.0
    # Could also use the viewport-page-sized approximation that
    # Gtk.ScrolledWindow uses internally:
    # https://git.gnome.org/browse/gtk+/tree/gtk/gtkscrolledwindow.c?h=gtk-3-14#n2416

    def __reset_delta_totals(self):
        self.__total_dx = 0.0
        self.__total_dy = 0.0

    def enter(self, doc, **kwds):
        self.__reset_delta_totals()
        return super(ScrollableModeMixin, self).enter(doc, **kwds)

    def button_press_cb(self, tdw, event):
        self.__reset_delta_totals()
        return super(ScrollableModeMixin, self).button_press_cb(tdw, event)

    def button_release_cb(self, tdw, event):
        self.__reset_delta_totals()
        return super(ScrollableModeMixin, self).button_release_cb(tdw, event)

    def scroll_cb(self, tdw, event):
        """Handles scroll-wheel events.

        Normal scroll wheel events: whichever of {panning, scrolling}
        the device is configured to do. With Ctrl or Alt: invert
        scrolling and zooming.

        With shift, if smooth scroll events are being sent, constrain
        the zoom or scroll in appropriate chunks.

        """
        doc = self.doc
        direction = event.direction
        dev_mon = doc.app.device_monitor
        dev = event.get_source_device()
        dev_settings = dev_mon.get_device_settings(dev)
        scroll_action = dev_settings.scroll

        # Invert scrolling and zooming if Ctrl or Alt is held
        import gui.device
        if event.state & (
                Gdk.ModifierType.MOD1_MASK | Gdk.ModifierType.CONTROL_MASK):
            if scroll_action == gui.device.ScrollAction.ZOOM:
                scroll_action = gui.device.ScrollAction.PAN
            elif scroll_action == gui.device.ScrollAction.PAN:
                scroll_action = gui.device.ScrollAction.ZOOM

        # Force incremental scrolling or zooming when shift is held.
        constrain_smooth = (event.state & Gdk.ModifierType.SHIFT_MASK)
        if direction == Gdk.ScrollDirection.SMOOTH:
            self.__total_dx += event.delta_x
            self.__total_dy += event.delta_y

        # Handle zooming (the old default)
        # We don't rotate any more though. Srsly, that was awful.
        if scroll_action == gui.device.ScrollAction.ZOOM:
            if direction == Gdk.ScrollDirection.SMOOTH:
                if constrain_smooth:
                    # Needs to work in an identical fashion to old-style
                    # zooming.
                    while self.__total_dy > 1:
                        self.__total_dy -= 1.0
                        doc.zoom(doc.ZOOM_OUTWARDS)
                    while self.__total_dy < -1:
                        self.__total_dy += 1.0
                        doc.zoom(doc.ZOOM_INWARDS)
                else:
                    # Smooth scroll zooming is intended to resemble what
                    # gui.viewmanip.ZoomViewMode does, minus the
                    # panning. In other words, simple zooming at the
                    # cursor.
                    dx = event.delta_x
                    dy = event.delta_y
                    dx *= self._PIXELS_PER_SMOOTH_SCROLL_UNIT
                    dy *= self._PIXELS_PER_SMOOTH_SCROLL_UNIT
                    # Don't pan: that's because the cursor generally does
                    # not move during scroll events.
                    # tdw.scroll(-dx, 0)  # not for now
                    dy *= -1
                    tdw.zoom(
                        math.exp(dy / 100.0),
                        center = (event.x, event.y),
                        ongoing = True,
                    )
                    tdw.renderer.update_cursor()
                    self.__reset_delta_totals()
                    # Need to send the notifications here if not
                    # callling doc methods.
                    # https://github.com/mypaint/mypaint/issues/313
                    doc.notify_view_changed()
            # Old-style zooming
            elif direction == Gdk.ScrollDirection.UP:
                doc.zoom(doc.ZOOM_INWARDS)
                self.__reset_delta_totals()
            elif direction == Gdk.ScrollDirection.DOWN:
                doc.zoom(doc.ZOOM_OUTWARDS)
                self.__reset_delta_totals()

        # Handle scroll panning.
        elif scroll_action == gui.device.ScrollAction.PAN:
            if direction == Gdk.ScrollDirection.SMOOTH:
                if constrain_smooth:
                    # Holding shift to constrain the pan works like
                    # discrete panning below.
                    while self.__total_dy > 1:
                        self.__total_dy -= 1.0
                        doc.pan(doc.PAN_DOWN)
                    while self.__total_dy < -1:
                        self.__total_dy += 1.0
                        doc.pan(doc.PAN_UP)
                    while self.__total_dx > 1:
                        self.__total_dx -= 1.0
                        doc.pan(doc.PAN_RIGHT)
                    while self.__total_dx < -1:
                        self.__total_dx += 1.0
                        doc.pan(doc.PAN_LEFT)
                else:
                    # Smooth panning is *nice*. It should work identically to
                    # gui.viewmanip.PanViewMode.
                    # No inertia here. Too many touchpads already
                    # emulate that, some by default and others not.
                    dx = event.delta_x
                    dy = event.delta_y
                    dx *= self._PIXELS_PER_SMOOTH_SCROLL_UNIT
                    dy *= self._PIXELS_PER_SMOOTH_SCROLL_UNIT
                    tdw.scroll(dx, dy, ongoing=True)
                    doc.notify_view_changed()
                    self.__reset_delta_totals()
            # Discrete panning.
            elif direction == Gdk.ScrollDirection.UP:
                doc.pan(doc.PAN_UP)
                self.__reset_delta_totals()
            elif direction == Gdk.ScrollDirection.DOWN:
                doc.pan(doc.PAN_DOWN)
                self.__reset_delta_totals()
            elif direction == Gdk.ScrollDirection.LEFT:
                doc.pan(doc.PAN_LEFT)
                self.__reset_delta_totals()
            elif direction == Gdk.ScrollDirection.RIGHT:
                doc.pan(doc.PAN_RIGHT)
                self.__reset_delta_totals()

        return super(ScrollableModeMixin, self).scroll_cb(tdw, event)


class PaintingModeOptionsWidgetBase (Gtk.Grid):
    """Base class for the options widget of a generic painting mode"""

    _COMMON_SETTINGS = [
        # TRANSLATORS:"Brush radius" for the options panel. Short.
        ('radius_logarithmic', _("Size:")),
        # TRANSLATORS:"Brush opacity" for the options panel. Short.
        ('opaque', _("Opaque:")),
        # TRANSLATORS:"Brush hardness/sharpness" for the options panel. Short.
        ('hardness', _("Sharp:")),
        # TRANSLATORS:"Additional pressure gain" for the options panel. Short.
        ('pressure_gain_log', _("Gain:")),
        # TRANSLATORS:"Paint Pigment Mode" for the options panel. Short.
        ('paint_mode', _("Pigment:")),
    ]

    def __init__(self):
        super(PaintingModeOptionsWidgetBase, self).__init__()
        self.set_row_spacing(6)
        self.set_column_spacing(6)
        from gui.application import get_app
        self.app = get_app()
        self.adjustable_settings = set()  #: What the reset button resets
        row = self.init_common_widgets(0)
        row = self.init_specialized_widgets(row)
        row = self.init_reset_widgets(row)

    def init_common_widgets(self, row):
        for cname, text in self._COMMON_SETTINGS:
            label = Gtk.Label()
            label.set_text(text)
            label.set_alignment(1.0, 0.5)
            label.set_hexpand(False)
            self.adjustable_settings.add(cname)
            adj = self.app.brush_adjustment[cname]
            scale = Gtk.Scale.new(Gtk.Orientation.HORIZONTAL, adj)
            scale.set_draw_value(False)
            scale.set_hexpand(True)
            self.attach(label, 0, row, 1, 1)
            self.attach(scale, 1, row, 1, 1)
            row += 1
        return row

    def init_specialized_widgets(self, row):
        return row

    def init_reset_widgets(self, row):
        align = Gtk.Alignment.new(0.5, 1.0, 1.0, 0.0)
        align.set_vexpand(True)
        self.attach(align, 0, row, 2, 1)
        button = Gtk.Button(label=_("Reset"))
        button.connect("clicked", self.reset_button_clicked_cb)
        align.add(button)
        row += 1
        return row

    def reset_button_clicked_cb(self, button):
        app = self.app
        bm = app.brushmanager
        parent_brush = bm.get_parent_brush(brushinfo=app.brush)
        if parent_brush:
            parent_binf = parent_brush.get_brushinfo()
            for cname in self.adjustable_settings:
                parent_value = parent_binf.get_base_value(cname)
                adj = self.app.brush_adjustment[cname]
                adj.set_value(parent_value)


class BrushworkModeMixin (InteractionMode):
    """Mixin for modes using brushes

    This mixin adds the ability to paint undoably to the current layer
    with proper atomicity and handling of checkpoints, and time-based
    automatic commits.

    Classes using this mixin should use `stroke_to()` to paint, and then
    may use the `brushwork_commit()` method to commit completed segments
    atomically to the command stack.  If a subclass needs greater
    control over new segments, `brushwork_begin()` can be used to start
    them recording.

    The `leave()` and `checkpoint()` methods defined here cooperatively
    commit all outstanding brushwork.
    """

    def __init__(self, **kwds):
        """Cooperative init (this mixin initializes some private fields)

        :param bool \*\*kwds: Passed through to other __init__s.

        """
        super(BrushworkModeMixin, self).__init__(**kwds)
        self.__first_begin = True
        self.__active_brushwork = {}  # {model: Brushwork}

    def brushwork_begin(self, model, description=None, abrupt=False,
                        layer=None):
        """Begins a new segment of active brushwork for a model

        :param Document model: The model to begin work on
        :param unicode description: Optional description of the work
        :param bool abrupt: Tail out/in abruptly with faked zero pressure.
        :param SimplePaintingLayer layer: explicit target layer.

        Any current segment of brushwork is committed, and a new segment
        is begun.

        Passing ``None`` for the description is suitable for freehand
        drawing modes.  This method will be called automatically with
        the default options by `stroke_to()` if needed, so not all
        mode classes will need to use it.

        The first segment of brushwork begun by a newly created
        BrushworkMode object always starts abruptly.
        The second and subsequent segments are assumed to be
        continuations by default. Set abrupt=True to break off any
        existing segment cleanly, and start the new segment cleanly.

        If an explicit target layer is used, it must be one that's
        guaranteed to persist for the lifetime of the current document
        model to prevent leaks.

        """
        # Commit any previous work for this model
        cmd = self.__active_brushwork.get(model)
        if cmd is not None:
            self.brushwork_commit(model, abrupt=abrupt)
        # New segment of brushwork
        if layer is None:
            layer_path = model.layer_stack.current_path
        else:
            layer_path = None
        cmd = lib.command.Brushwork(
            model,
            layer_path=layer_path,
            description=description,
            abrupt_start=(abrupt or self.__first_begin),
            layer=layer,
        )
        self.__first_begin = False
        cmd.__last_pos = None
        self.__active_brushwork[model] = cmd

    def brushwork_commit(self, model, abrupt=False):
        """Commits any active brushwork for a model to the command stack

        :param Document model: The model to commit work to
        :param bool abrupt: End with a faked zero pressure "stroke_to()"

        This only makes a new entry on the command stack if
        the currently active brushwork segment made
        any changes to the model.

        See also `brushwork_rollback()`.
        """
        cmd = self.__active_brushwork.pop(model, None)
        if cmd is None:
            return
        if abrupt and cmd.__last_pos is not None:
            x, y, xtilt, ytilt, viewzoom, viewrotation, barrel_rotation = cmd.__last_pos
            pressure = 0.0
            dtime = 0.0
            viewzoom = self.doc.tdw.scale
            viewrotation = self.doc.tdw.rotation
            barrel_rotation = 0.0
            cmd.stroke_to(dtime, x, y, pressure, xtilt, ytilt,
                          viewzoom, viewrotation, barrel_rotation)
        changed = cmd.stop_recording(revert=False)
        if changed:
            model.do(cmd)

    def brushwork_rollback(self, model):
        """Rolls back any active brushwork for a model

        :param Document model: The model to roll back

        This restores the model's appearance and state to how it was
        when the current segment of brushwork started.
        For input patterns where this makes sense,
        your calls to `stroke_to()` should have ``auto_split=False``.

        See also `brushwork_commit()`.
        """
        cmd = self.__active_brushwork.pop(model, None)
        if cmd is None:
            return
        cmd.stop_recording(revert=True)

    def brushwork_commit_all(self, abrupt=False):
        """Commits all active brushwork"""
        for model in list(self.__active_brushwork.keys()):
            self.brushwork_commit(model, abrupt=abrupt)

    def brushwork_rollback_all(self):
        """Rolls back all active brushwork"""
        for model in list(self.__active_brushwork.keys()):
            self.brushwork_rollback(model)

    def stroke_to(self, model, dtime, x, y, pressure, xtilt, ytilt,
                  viewzoom, viewrotation, barrel_rotation,
                  auto_split=True, layer=None):
        """Feeds an updated stroke position to the brush engine

        :param Document model: model on which to paint
        :param float dtime: Seconds since the last call to this method
        :param float x: Document X position update
        :param float y: Document Y position update
        :param float pressure: Pressure, ranging from 0.0 to 1.0
        :param float xtilt: X-axis tilt, ranging from -1.0 to 1.0
        :param float ytilt: Y-axis tilt, ranging from -1.0 to 1.0
        :param viewzoom: The view's current zoom level, [0, 64]
        :param viewrotation: The view's current rotation, [-180.0, 180.0]
        :param float barrel_rotation: Stylus barrel rotation, [0.0 to 1.0]
        :param bool auto_split: Split ongoing brushwork if due
        :param SimplePaintingLayer layer: explicit target layer

        During normal operation, successive calls to `stroke_to()` record
        an ongoing sequence of `lib.command.Brushwork` commands on the
        undo stack, stopping and committing the currently recording
        command when it becomes due.

        The explicit target layer is intended for simple painting modes
        operating on out-of-tree layers which rely on stroke_to()
        automatically calling brushwork_begin(). Normally the currently
        selected layer is used as the target layer for each new segment
        of brushwork.

        """
        cmd = self.__active_brushwork.get(model, None)
        desc0 = None
        if auto_split and cmd and cmd.split_due:
            desc0 = cmd.description  # retain for the next cmd
            self.brushwork_commit(model, abrupt=False)
            assert model not in self.__active_brushwork
            cmd = None
        if not cmd:
            self.brushwork_begin(
                model,
                description=desc0,
                abrupt=False,
                layer=layer,
            )
            cmd = self.__active_brushwork[model]
        cmd.stroke_to(dtime, x, y, pressure, xtilt, ytilt,
                      viewzoom, viewrotation, barrel_rotation)
        cmd.__last_pos = (x, y, xtilt, ytilt, viewzoom,
                          viewrotation, barrel_rotation)

    def leave(self, **kwds):
        """Leave mode, committing outstanding brushwork as necessary

        The leave action defined here is careful to tail off strokes
        cleanly: certain subclasses are geared towards fast capture of
        data and queued delivery of stroke information, so we have to
        reset the brush engine's idea of pressure fast. If we don't, an
        interrupted queued stroke can result in a *huge* sequence of
        dabs from the last processed position to wherever the cursor is
        right now.

        This leave() knows about the mode stack, and only commits if it
        knows it isn't still stacked. That's to allow temporary view
        manipulation modes to work without disrupting `gui.inktool`'s
        mode, which normally has a lot pending.

        """
        logger.debug("BrushworkModeMixin: leave()")
        # FIXME: The mode stack should be telling enter() and leave()
        # FIXME: whether this is an initial/final call.
        # FIXME: Stack state tracking should be unnecessary inside mode objs.
        still_stacked = False
        for mode in self.doc.modes:
            if mode is self:
                still_stacked = True
                break
        if not still_stacked:
            self.brushwork_commit_all(abrupt=True)
        super(BrushworkModeMixin, self).leave(**kwds)

    def checkpoint(self, **kwargs):
        """Commit any outstanding brushwork

        Like `leave()`, this commits the currently recording Brushwork
        command for each known model; however we do not attempt to tail
        off brushstrokes cleanly because that would make Freehand mode
        discontinuous when the user changes the brush color.

        """
        logger.debug("BrushworkModeMixin: checkpoint()")
        super(BrushworkModeMixin, self).checkpoint(**kwargs)
        self.brushwork_commit_all(abrupt=False)


class SingleClickMode (InteractionMode):
    """Base class for non-drag (single click) modes"""

    #: The cursor to use when entering the mode
    cursor = None

    def __init__(self, ignore_modifiers=False, **kwds):
        super(SingleClickMode, self).__init__(**kwds)
        self._button_pressed = None

    def enter(self, doc, **kwds):
        super(SingleClickMode, self).enter(doc, **kwds)
        assert self.doc is not None
        self.doc.tdw.set_override_cursor(self.cursor)
        if self.cursor is None:
            self.cursor = self.doc.app.cursors.get_action_cursor(
                    self.ACTION_NAME, gui.cursor.Name.ARROW)

    def leave(self, **kwds):
        if self.doc is not None:
            self.doc.tdw.set_override_cursor(None)
        super(SingleClickMode, self).leave(**kwds)

    def button_press_cb(self, tdw, event):
        if event.button == 1 and event.type == Gdk.EventType.BUTTON_PRESS:
            self._button_pressed = 1
            return False
        else:
            return super(SingleClickMode, self).button_press_cb(tdw, event)

    def button_release_cb(self, tdw, event):
        if event.button == self._button_pressed:
            self._button_pressed = None
            self.clicked_cb(tdw, event)
            return False
        else:
            return super(SingleClickMode, self).button_press_cb(tdw, event)

    def clicked_cb(self, tdw, event):
        assert not hasattr(super(SingleClickMode, self), "clicked_cb")


class DragMode (InteractionMode):
    """Base class for drag activities.

    Dragm modes can be entered when the button is pressed, or not yet
    pressed.  If the button is pressed when the mode is entered. the
    initial position will be determined from the first motion event.

    Drag modes are normally "spring-loaded", meaning that when a drag
    mode is first entered, it remembers which modifier keys were held
    down at that time. When these keys are released, the mode will exit.

    """

    inactive_cursor = None
    active_cursor = None

    #: If true, exit mode when initial modifiers are released
    SPRING_LOADED = True

    def __init__(self, ignore_modifiers=False, **kwds):
        """Construct, possibly ignoring initial modifiers.

        :param ignore_modifiers: If True, ignore initial modifier keys.

        Drag modes can be instructed to ignore the initial set of
        modifiers when they're entered. This is appropriate when the
        mode is being entered in response to a keyboard shortcut.
        Modifiers don't mean the same thing for keyboard shortcuts.
        Conversely, toolbar buttons and mode-switching via pointer
        buttons should use the default behaviour.

        In practice, it's not quite so clear cut. Instead we have
        keyboard-friendly "Flip*" actions which allow the mode to be
        toggled off with a second press.  These actions use the
        `ignore_modifiers` behaviour, and coexist with a secondary layer
        of radioactions which don't do this, but which reflect the state
        prettily.

        """
        super(DragMode, self).__init__(**kwds)
        self._tdw_grab_broken_conninfo = None
        self._reset_drag_state()
        self.initial_modifiers = None
        #: Ignore the initial modifiers (FIXME: bad name, maybe not public?)
        self.ignore_modifiers = ignore_modifiers

    def _reset_drag_state(self):
        self.last_x = None
        self.last_y = None
        self.start_x = None
        self.start_y = None
        self._start_keyval = None
        self._start_button = None
        self._grab_widget = None
        if self._tdw_grab_broken_conninfo is not None:
            tdw, connid = self._tdw_grab_broken_conninfo
            tdw.disconnect(connid)
            self._tdw_grab_broken_conninfo = None

    def _stop_drag(self, t=Gdk.CURRENT_TIME):
        # Stops any active drag, calls drag_stop_cb(), and cleans up.
        if not self.in_drag:
            return
        tdw = self._grab_widget
        tdw.grab_remove()
        Gdk.keyboard_ungrab(t)
        Gdk.pointer_ungrab(t)
        self._grab_widget = None
        self.drag_stop_cb(tdw)
        self._reset_drag_state()

    def _start_drag(self, tdw, event):
        # Attempt to start a new drag, calling drag_start_cb() if successful.
        if self.in_drag:
            return
        if hasattr(event, "x"):
            self.start_x = event.x
            self.start_y = event.y
        else:
            last_t, last_x, last_y = self.doc.get_last_event_info(tdw)
            self.start_x = last_x
            self.start_y = last_y
        tdw_window = tdw.get_window()
        event_mask = (Gdk.EventMask.BUTTON_PRESS_MASK |
                      Gdk.EventMask.BUTTON_RELEASE_MASK |
                      Gdk.EventMask.POINTER_MOTION_MASK)
        cursor = self.active_cursor
        if cursor is None:
            cursor = self.inactive_cursor

        # Grab the pointer
        grab_status = Gdk.pointer_grab(tdw_window, False, event_mask, None,
                                       cursor, event.time)
        if grab_status != Gdk.GrabStatus.SUCCESS:
            logger.warning("pointer grab failed: %r", grab_status)
            logger.debug("gdk_pointer_is_grabbed(): %r",
                         Gdk.pointer_is_grabbed())
            # There seems to be a race condition between this grab under
            # PyGTK/GTK2 and some other grab - possibly just the implicit grabs
            # on color selectors: https://gna.org/bugs/?20068 Only pointer
            # events are affected, and PyGI+GTK3 is unaffected.
            #
            # It's probably safest to exit the mode and not start the drag.
            # This condition should be rare enough for this to be a valid
            # approach: the irritation of having to click again to do something
            # should be far less than that of getting "stuck" in a drag.
            self._bailout()

            # Sometimes a pointer ungrab is needed even though the grab
            # apparently failed to avoid the UI partially "locking up" with the
            # stylus (and only the stylus). Happens when WMs like Xfwm
            # intercept an <Alt>Button combination for window management
            # purposes. Results in Gdk.GrabStatus.ALREADY_GRABBED, but this
            # line is necessary to avoid the rest of the UI becoming
            # unresponsive even though the canvas can be drawn on with
            # the stylus. Are we cancelling an implicit grab here, and why
            # is it device specific?
            Gdk.pointer_ungrab(event.time)
            return

        # We managed to establish a grab, so watch for it being broken.
        # This signal is disconnected when the mode leaves.
        connid = tdw.connect("grab-broken-event", self._tdw_grab_broken_cb)
        self._tdw_grab_broken_conninfo = (tdw, connid)

        # Grab the keyboard too, to be certain of getting the key release event
        # for a spacebar drag.
        grab_status = Gdk.keyboard_grab(tdw_window, False, event.time)
        if grab_status != Gdk.GrabStatus.SUCCESS:
            logger.warning("Keyboard grab failed: %r", grab_status)
            self._bailout()
            Gdk.pointer_ungrab(event.time)
            return

        # GTK too...
        tdw.grab_add()
        self._grab_widget = tdw

        # Drag has started, perform whatever action the mode needs.
        self.drag_start_cb(tdw, event)

    def _bailout(self):
        """Attempt to exit this mode safely, via an idle routine

        The actual task is handled by an idle callback to make this
        method safe to call during a mode's enter() or leave() methods.
        Modes on top of the one requesting bailout will also be ejected.

        """
        from gui.application import get_app
        app = get_app()
        if self not in app.doc.modes:
            logger.debug(
                "bailout: cannot bail out of %r: "
                "mode is not in the mode stack",
                self,
            )
            return
        logger.debug("bailout: starting idler to safely bail out of %r", self)
        GLib.idle_add(self._bailout_idle_cb, app.doc.modes)

    def _bailout_idle_cb(self, modestack):
        """Bail out of this mode if it's anywhere in the mode stack"""
        while self in modestack:
            logger.debug("bailout idler: leaving %r", modestack.top)
            modestack.pop()
        logger.debug("bailout idler: done")
        return False

    def _tdw_grab_broken_cb(self, tdw, event):
        # Cede control as cleanly as possible if something else grabs either
        # the keyboard or the pointer while a grab is active.
        # One possible cause for https://gna.org/bugs/?20333
        logger.debug("grab-broken-event on %r", tdw)
        logger.debug(" send_event  : %r", event.send_event)
        logger.debug(" keyboard    : %r", event.keyboard)
        logger.debug(" implicit    : %r", event.implicit)
        logger.debug(" grab_window : %r", event.grab_window)
        self._bailout()
        return True

    @property
    def in_drag(self):
        return self._grab_widget is not None

    def enter(self, doc, **kwds):
        """Enter the mode, recording the held modifier keys the 1st time

        The attribute `self.initial_modifiers` is set the first time the
        mode is entered.

        """
        super(DragMode, self).enter(doc, **kwds)
        assert self.doc is not None
        if self.inactive_cursor is None:
            # some children might override self.inactive_cursor as read-only
            try:
                self.inactive_cursor = self.doc.app.cursors. \
                        get_action_cursor(self.ACTION_NAME,
                                          gui.cursor.Name.ARROW)
            except AttributeError:
                pass
        self.doc.tdw.set_override_cursor(self.inactive_cursor)

        if self.SPRING_LOADED:
            if self.ignore_modifiers:
                self.initial_modifiers = 0
                return
            old_modifiers = getattr(self, "initial_modifiers", None)
            if old_modifiers is not None:
                # Re-entering due to an overlying mode being popped
                if old_modifiers != 0:
                    # This mode started with modifiers held
                    modifiers = self.current_modifiers()
                    if (modifiers & old_modifiers) == 0:
                        # But none of them are held any more,
                        # so queue a further pop.
                        GLib.idle_add(self.__pop_modestack_idle_cb)
            else:
                # This mode is being entered for the first time;
                # record modifiers
                modifiers = self.current_modifiers()
                self.initial_modifiers = self.current_modifiers()

    def __pop_modestack_idle_cb(self):
        # Pop the mode stack when this mode is re-entered but has to leave
        # straight away because its modifiers are no longer held. Doing it in
        # an idle function avoids confusing the derived class's enter() method:
        # a leave() during an enter() would be strange.
        if self.initial_modifiers is not None:
            if self.doc and (self is self.doc.modes.top):
                self.doc.modes.pop()
        return False

    def leave(self, **kwds):
        self._stop_drag()
        if self.doc is not None:
            self.doc.tdw.set_override_cursor(None)
        super(DragMode, self).leave(**kwds)

    def button_press_cb(self, tdw, event):
        if event.type == Gdk.EventType.BUTTON_PRESS:
            if self.in_drag:
                if self._start_button is None:
                    # Doing this allows single clicks to exit keyboard
                    # initiated drags, e.g. those forced when handling a
                    # keyboard event somewhere else.
                    self._start_button = event.button
            else:
                self._start_drag(tdw, event)
                if self.in_drag:
                    # Grab succeeded
                    self.last_x = event.x
                    self.last_y = event.y
                    self._start_button = event.button
        return super(DragMode, self).button_press_cb(tdw, event)

    def button_release_cb(self, tdw, event):
        if self.in_drag:
            if event.button == self._start_button:
                self._stop_drag()
        return super(DragMode, self).button_release_cb(tdw, event)

    def motion_notify_cb(self, tdw, event):
        # We might be here because an Action manipulated the modes stack
        # but if that's the case then we should wait for a button or
        # a keypress to initiate the drag.
        if self.in_drag:
            if self.last_x is not None:
                dx = event.x - self.last_x
                dy = event.y - self.last_y
                self.drag_update_cb(tdw, event, dx, dy)
            self.last_x = event.x
            self.last_y = event.y
            return True
        # Fall through to other behavioral mixins, just in case
        return super(DragMode, self).motion_notify_cb(tdw, event)

    def key_press_cb(self, win, tdw, event):
        if self.in_drag:
            # Eat keypresses in the middle of a drag no matter how
            # it was started.
            return True
        elif event.keyval == Gdk.KEY_space:
            # Start drags on space
            if event.keyval != self._start_keyval:
                self._start_keyval = event.keyval
                self._start_drag(tdw, event)
            return True
        # Fall through to other behavioral mixins
        return super(DragMode, self).key_press_cb(win, tdw, event)

    def key_release_cb(self, win, tdw, event):
        if self.in_drag:
            if event.keyval == self._start_keyval:
                self._stop_drag()
                self._start_keyval = None
            return True

        if self.SPRING_LOADED:
            if event.is_modifier and self.in_drag:
                return False
            if self.initial_modifiers:
                modifiers = self.current_modifiers()
                if modifiers & self.initial_modifiers == 0:
                    if self is self.doc.modes.top:
                        self.doc.modes.pop()
                    return True

        # Fall through to other behavioral mixins
        return super(DragMode, self).key_release_cb(win, tdw, event)


class OneshotDragMode (DragMode):
    """Drag modes that can exit immediately when the drag stops

    These are utility modes which allow the user to do quick, simple
    tasks with the canvas like pick a color from it or pan the view.
    """
    def __init__(self, unmodified_persist=True, temporary_activation=True,
                 **kwargs):
        """
        :param bool unmodified_persist: Stay active if entered without modkeys
        :param bool \*\*kwargs: Passed through to other __init__s.

        If unmodified_persist is true, and drag mode is spring-loaded, the
        tool will stay active if no modifiers were held initially. This means
        tools will not deselect themselves after one use if activated from,
        say, the toolbar.
        """
        DragMode.__init__(self)
        self.unmodified_persist = unmodified_persist
        self.temporary_activation = temporary_activation

    def stackable_on(self, mode):
        """Oneshot modes return to the mode the user came from on exit"""
        return not isinstance(mode, OneshotDragMode)

    def drag_stop_cb(self, tdw):
        if not hasattr(self, "initial_modifiers"):
            # Always exit at the end of a drag if not spring-loaded.
            if self is self.doc.modes.top:
                self.doc.modes.pop()
        elif self.initial_modifiers != 0:
            # If started with modifiers, keeping the modifiers held keeps
            # spring-loaded modes active. If not, exit the mode.
            if (self.initial_modifiers & self.current_modifiers()) == 0:
                if self is self.doc.modes.top:
                    self.doc.modes.pop()
        else:
            # No modifiers were held when this mode was entered.
            if self.temporary_activation or (not self.unmodified_persist):
                if self is self.doc.modes.top:
                    self.doc.modes.pop()
        return super(OneshotDragMode, self).drag_stop_cb(tdw)


## Mode stack


class _NullMode (InteractionMode):
    """A mode that does nothing (placeholder only)"""


class ModeStack (object):
    """A stack of InteractionModes. The top mode is the active one.

    Mode stacks can never be empty. If the final element is popped, it
    will be replaced with a new instance of its ``default_mode_class``.

    """

    def __init__(self, doc):
        """Initialize for a particular controller

        :param doc: Controller instance
        :type doc: CanvasController

        The main MyPaint app uses an instance of `gui.document.Document`
        as `doc`. Simpler drawing canvases can use a basic
        CanvasController and a simpler `default_mode_class`.
        """
        object.__init__(self)
        self._stack = []
        self._doc = doc
        self._syncing_pending_changes = False
        if hasattr(doc, "model"):
            doc.model.sync_pending_changes += self._sync_pending_changes_cb
        #: Class to instantiate if stack is empty: callable with 0 args.
        self.default_mode_class = _NullMode

    def _sync_pending_changes_cb(self, model, **kwargs):
        """Syncs pending changes with the model

        :param Document model: the requesting model
        :param \*\*kwargs: passed through to checkpoint()

        This issues a `checkpoint()` on the current InteractionMode.
        """
        if self._syncing_pending_changes:
            return
        self._syncing_pending_changes = True
        self.top.checkpoint(**kwargs)
        self._syncing_pending_changes = False

    @event
    def changed(self, old, new):
        """Event: emitted when the active mode changes

        :param old: The previous active mode
        :param new: The new `top` (current) mode

        This event is emitted after the ``enter()`` method of the new
        mode has been called, and therefore after the ``leave()`` of the
        old mode too. On occasion, the old mode may be null.

        Context-aware pushes call this only once, with the old active
        and newly active mode only regardless of how many modes were
        skipped.

        """

    @property
    def top(self):
        """The top node on the stack"""
        # Perhaps rename to "active()"?
        new_mode = self._check()
        if new_mode is not None:
            new_mode.enter(doc=self._doc)
            self.changed(None, new_mode)
        return self._stack[-1]

    def context_push(self, mode):
        """Context-aware push.

        :param mode: mode to be stacked and made active
        :type mode: `InteractionMode`

        Stacks a mode onto the topmost element in the stack it is compatible
        with, as determined by its ``stackable_on()`` method. Incompatible
        top modes are popped one by one until either a compatible mode is
        found, or the stack is emptied, then the new mode is pushed.

        """
        # Pop until the stack is empty, or the top mode is compatible
        old_mode = None
        if len(self._stack) > 0:
            old_mode = self._stack[-1]
        while len(self._stack) > 0:
            if mode.stackable_on(self._stack[-1]):
                break
            self._stack.pop(-1).leave()
            if len(self._stack) > 0:
                self._stack[-1].enter(doc=self._doc)
        # Stack on top of any remaining compatible mode
        if len(self._stack) > 0:
            self._stack[-1].leave()
        self._stack.append(mode)
        mode.enter(doc=self._doc)
        self.changed(old=old_mode, new=mode)

    def pop(self):
        """Pops a mode, leaving the old top mode and entering the exposed top.
        """
        old_mode = None
        if len(self._stack) > 0:
            old_mode = self._stack.pop(-1)
            old_mode.leave()
        top_mode = self._check()
        if top_mode is None:
            top_mode = self._stack[-1]
        # No need to checkpoint user activity here: leave() was already called
        top_mode.enter(doc=self._doc)
        self.changed(old=old_mode, new=top_mode)

    def push(self, mode):
        """Pushes a mode, and enters it.

        :param mode: Mode to be stacked and made active
        :type mode: InteractionMode
        """
        old_mode = None
        if len(self._stack) > 0:
            old_mode = self._stack[-1]
            old_mode.leave()
        # No need to checkpoint user activity here: leave() was already called
        self._stack.append(mode)
        mode.enter(doc=self._doc)
        self.changed(old=old_mode, new=mode)

    def reset(self, replacement=None):
        """Clears the stack, popping the final element and replacing it.

        :param replacement: Optional mode to go on top of the cleared stack.
        :type replacement: `InteractionMode`.

        """
        old_top_mode = None
        if len(self._stack) > 0:
            old_top_mode = self._stack[-1]
        while len(self._stack) > 0:
            old_mode = self._stack.pop(-1)
            old_mode.leave()
            if len(self._stack) > 0:
                self._stack[-1].enter(doc=self._doc)
        top_mode = self._check(replacement)
        assert top_mode is not None
        self.changed(old=old_top_mode, new=top_mode)

    def pop_to_behaviour(self, flags):
        """Keeps popping the stack until a node that matches the flags is found.
        If the stack does not contain such a node, you will simply end up with
        an empty ModeStack.

        :param flags: Descriptors of the node you want.
        :type flags: `Behavior`.

        By "empty ModeStack" I mean a ModeStack with a single
        ``default_mode_class`` instance, as usual.
        """
        while self.top.pointer_behavior & flags == 0:
            if len(self._stack) == 1:
                self.pop()
                return
            self.pop()

    def _check(self, replacement=None):
        """Ensures that the stack is non-empty

        :param replacement: Optional replacement mode instance.
        :type replacement: `InteractionMode`.

        Returns the new top mode if one was pushed.
        """
        if len(self._stack) > 0:
            return None
        if replacement is not None:
            mode = replacement
        else:
            mode = self.default_mode_class()
        self._stack.append(mode)
        mode.enter(doc=self._doc)
        return mode

    def __repr__(self):
        """Plain-text representation."""
        s = '<ModeStack ['
        s += ", ".join([m.__class__.__name__ for m in self._stack])
        s += ']>'
        return s

    def __len__(self):
        """Returns the number of modes on the stack."""
        return len(self._stack)

    def __nonzero__(self):
        """Mode stacks never test false, regardless of length."""
        return True

    def __iter__(self):
        for mode in self._stack:
            yield mode



# This file is part of MyPaint.
# Copyright (C) 2012-2018 by the MyPaint Development Team.
# Copyright (C) 2015 by ShadowKyogre <shadowkyogre@aim.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.


"""Axis-aligned planar slice of an HSV color cube, and a depth slider.
"""

from __future__ import division, print_function
import math
from gettext import gettext as _

from lib.gibindings import Gtk
import cairo

from .util import clamp
from .util import draw_marker_circle
from lib.color import RGBColor, HSVColor
from .bases import IconRenderable
from .adjbases import ColorAdjusterWidget
from .adjbases import ColorAdjuster
from .adjbases import IconRenderableColorAdjusterWidget
from .adjbases import HueSaturationWheelAdjuster
from .combined import CombinedAdjusterPage

from lib.pycompat import xrange


class HSVSquarePage (CombinedAdjusterPage, IconRenderable):
    """Hue ring and Sat+Val square: page for `CombinedAdjuster`."""

    def __init__(self):
        self._faces = ['h', 's', 'v']
        table = Gtk.Table(n_rows=1, n_columns=1)

        xopts = Gtk.AttachOptions.FILL | Gtk.AttachOptions.EXPAND
        yopts = Gtk.AttachOptions.FILL | Gtk.AttachOptions.EXPAND

        self.__adj = HSVSquare()

        table.attach(self.__adj, 0, 1, 0, 1, xopts, yopts, 3, 3)
        self.__table = table
        self.__adj._update_tooltips()

    @classmethod
    def get_page_icon_name(self):
        return 'mypaint-tool-hsvsquare'

    @classmethod
    def get_page_title(self):
        return _('HSV Square')

    @classmethod
    def get_page_description(self):
        return _("An HSV Square which can be rotated to show different hues.")

    def get_page_widget(self):
        return self.__table

    def set_color_manager(self, manager):
        ColorAdjuster.set_color_manager(self, manager)
        self.__adj.set_color_manager(manager)

    def render_as_icon(self, cr, size):
        """Renders as an icon into a Cairo context.
        """
        # Strategy: construct tmp R,G,B sliders with a color that shows off
        # their primary a bit. Render carefully (might need special handling
        # for the 16px size).
        from adjbases import ColorManager
        mgr = ColorManager(prefs={}, datapath=".")
        mgr.set_color(RGBColor(0.3, 0.3, 0.4))
        ring_adj = _HSVSquareOuterRing(self)
        ring_adj.set_color_manager(mgr)
        square_adj = _HSVSquareInnerSquare(self)
        square_adj.set_color_manager(mgr)
        if size <= 16:
            cr.save()
            ring_adj.render_background_cb(cr, wd=16, ht=16)
            cr.translate(-6, -6)
            square_adj.render_background_cb(cr, wd=12, ht=12)
            cr.restore()
        else:
            cr.save()
            square_offset = int(size/5.0 * 1.6)
            square_dim = int(size * 0.64)
            ring_adj.render_background_cb(cr, wd=size, ht=size)
            # do minor rounding adjustments for hsvsquare icons at this size
            if size == 24:
                cr.translate(-1, -1)
                square_dim += 1
            cr.translate(-square_offset, -square_offset)
            square_adj.render_background_cb(cr, wd=square_dim, ht=square_dim)
            cr.restore()
        ring_adj.set_color_manager(None)
        square_adj.set_color_manager(None)


class HSVSquare(Gtk.VBox, ColorAdjuster):
    """Combined Sat+Val square and Hue ring color adjuster"""

    __gtype_name__ = 'HSVSquare'

    def __init__(self):
        super(HSVSquare, self).__init__()
        self._faces = ['h', 's', 'v']

        self.__square = _HSVSquareInnerSquare(self)
        self.__ring = _HSVSquareOuterRing(self)

        s_align = Gtk.Alignment(
            xalign=0.5, yalign=0.5,
            xscale=0.54, yscale=0.54,
        )
        plz_be_square = Gtk.AspectFrame()
        plz_be_square.set_shadow_type(Gtk.ShadowType.NONE)
        s_align.add(plz_be_square)
        plz_be_square.add(self.__square)
        self.__ring.add(s_align)
        self.pack_start(self.__ring, True, True, 0)

    def set_color_manager(self, manager):
        super(HSVSquare, self).set_color_manager(manager)
        self.__square.set_color_manager(manager)
        self.__ring.set_color_manager(manager)

    def _update_tooltips(self):
        self.__ring.set_tooltip_text(_("HSV Hue"))
        self.__square.set_tooltip_text(_("HSV Saturation and Value"))


class _HSVSquareOuterRing (HueSaturationWheelAdjuster):
    """Outer color ring"""

    vertical = True
    samples = 4

    def __init__(self, cube):
        HueSaturationWheelAdjuster.__init__(self)
        self.__cube = cube

    def get_pos_for_color(self, col):
        nr, ntheta = self.get_normalized_polar_pos_for_color(col)
        mgr = self.get_color_manager()
        if mgr:
            ntheta = mgr.distort_hue(ntheta)
        nr **= 1.0/self.SAT_GAMMA
        alloc = self.get_allocation()
        wd, ht = alloc.width, alloc.height
        radius = self.get_radius(wd, ht, self.BORDER_WIDTH)
        cx, cy = self.get_center(wd, ht)
        r = radius * clamp(nr, 0, 1)
        t = clamp(ntheta, 0, 1) * 2 * math.pi
        x = int(cx + r*math.cos(t)) + 0.5
        y = int(cy + r*math.sin(t)) + 0.5
        return x, y

    def get_color_at_position(self, x, y):
        """Gets the color at a position, for `ColorAdjusterWidget` impls.
        """
        alloc = self.get_allocation()
        cx, cy = self.get_center(alloc=alloc)
        # Normalized radius
        r = math.sqrt((x-cx)**2 + (y-cy)**2)
        radius = self.get_radius(alloc=alloc)
        if r > radius:
            r = radius
        r /= radius
        r **= self.SAT_GAMMA
        # Normalized polar angle
        theta = 1.25 - (math.atan2(x-cx, y-cy) / (2*math.pi))
        while theta <= 0:
            theta += 1.0
        theta %= 1.0
        mgr = self.get_color_manager()
        if mgr:
            theta = mgr.undistort_hue(theta)
        return self.color_at_normalized_polar_pos(r, theta)

    def get_normalized_polar_pos_for_color(self, col):
        col = HSVColor(color=col)
        return col.s, col.h

    def color_at_normalized_polar_pos(self, r, theta):
        col = HSVColor(color=self.get_managed_color())
        col.h = theta
        return col

    def get_background_validity(self):
        col = HSVColor(color=self.get_managed_color())
        f0, f1, f2 = self.__cube._faces
        return f0, getattr(col, f0)

    def render_background_cb(self, cr, wd, ht, icon_border=None):
        """Renders the offscreen bg, for `ColorAdjusterWidget` impls.
        """
        cr.save()

        border = icon_border
        if border is None:
            border = self.BORDER_WIDTH
        radius = self.get_radius(wd, ht, border)

        steps = self.HUE_SLICES

        # Move to the centre
        cx, cy = self.get_center(wd, ht)
        cr.translate(cx, cy)

        # Clip, for a slight speedup
        cr.arc(0, 0, radius+border, 0, 2*math.pi)
        cr.clip()

        # Tangoesque outer border
        cr.set_line_width(self.OUTLINE_WIDTH)
        cr.arc(0, 0, radius, 0, 2*math.pi)
        cr.set_source_rgba(*self.OUTLINE_RGBA)
        cr.stroke()

        # Each slice in turn
        cr.save()
        cr.set_line_width(1.0)
        cr.set_line_join(cairo.LINE_JOIN_ROUND)
        step_angle = 2.0*math.pi/steps
        mgr = self.get_color_manager()

        for ih in xrange(steps+1):  # overshoot by 1, no solid bit for final
            h = ih / steps
            if mgr:
                h = mgr.undistort_hue(h)
            edge_col = self.color_at_normalized_polar_pos(1.0, h)
            edge_col.s = 1.0
            edge_col.v = 1.0
            rgb = edge_col.get_rgb()

            if ih > 0:
                # Backwards gradient
                cr.arc_negative(0, 0, radius, 0, -step_angle)
                x, y = cr.get_current_point()
                cr.line_to(0, 0)
                cr.close_path()
                lg = cairo.LinearGradient(radius, 0, (x + radius) / 2, y)
                lg.add_color_stop_rgba(0, rgb[0], rgb[1], rgb[2], 1.0)
                lg.add_color_stop_rgba(1, rgb[0], rgb[1], rgb[2], 0.0)
                cr.set_source(lg)
                cr.fill()

            if ih < steps:
                # Forward solid
                cr.arc(0, 0, radius, 0, step_angle)
                x, y = cr.get_current_point()
                cr.line_to(0, 0)
                cr.close_path()
                cr.set_source_rgb(*rgb)
                cr.stroke_preserve()
                cr.fill()
            cr.rotate(step_angle)

        cr.restore()

        # Tangoesque inner border
        cr.set_source_rgba(*self.EDGE_HIGHLIGHT_RGBA)
        cr.set_line_width(self.EDGE_HIGHLIGHT_WIDTH)
        cr.arc(0, 0, radius, 0, 2*math.pi)
        cr.stroke()

        cr.set_line_width(self.OUTLINE_WIDTH)
        cr.arc(0, 0, radius*0.8, 0, 2*math.pi)
        cr.set_source_rgba(0, 0, 0, 1)
        cr.set_operator(cairo.OPERATOR_DEST_OUT)
        cr.fill()
        cr.set_operator(cairo.OPERATOR_OVER)
        cr.set_source_rgba(*self.OUTLINE_RGBA)
        cr.stroke()

        cr.set_source_rgba(*self.EDGE_HIGHLIGHT_RGBA)
        cr.set_line_width(self.EDGE_HIGHLIGHT_WIDTH)
        cr.arc(0, 0, radius*0.8, 0, 2*math.pi)
        cr.stroke()

    def paint_foreground_cb(self, cr, wd, ht):
        """Fg marker painting, for `ColorAdjusterWidget` impls.
        """
        col = HSVColor(color=self.get_managed_color())
        col.s = 1.0
        radius = self.get_radius(wd, ht, self.BORDER_WIDTH)
        cx = int(wd // 2)
        cy = int(ht // 2)
        cr.arc(cx, cy, radius+0.5, 0, 2*math.pi)
        cr.clip()
        x, y = self.get_pos_for_color(col)
        col.s = 0.70
        ex, ey = self.get_pos_for_color(col)

        cr.set_line_width(5)
        cr.move_to(x, y)
        cr.line_to(ex, ey)
        cr.set_source_rgb(0, 0, 0)
        cr.stroke_preserve()

        cr.set_source_rgb(1, 1, 1)
        cr.set_line_width(3.5)
        cr.stroke_preserve()

        cr.set_source_rgb(*col.get_rgb())
        cr.set_line_width(0.25)
        cr.stroke()


class _HSVSquareInnerSquare (IconRenderableColorAdjusterWidget):
    """Inner saturation & value square"""

    def __init__(self, cube):
        ColorAdjusterWidget.__init__(self)
        self.__cube = cube
        self.connect('button-press-event', self.stop_fallthrough)

    def stop_fallthrough(self, widget, event):
        return True

    def __get_faces(self):
        f1 = self.__cube._faces[1]
        f2 = self.__cube._faces[2]
        if f2 == 'h':
            f1, f2 = f2, f1
        return f1, f2

    def render_background_cb(self, cr, wd, ht, icon_border=None):
        col = HSVColor(color=self.get_managed_color())
        b = icon_border
        if b is None:
            b = self.BORDER_WIDTH
        eff_wd = int(wd - 2*b)
        eff_ht = int(ht - 2*b)
        f1, f2 = self.__get_faces()

        step = max(1, int(eff_wd // 128))

        rect_x, rect_y = int(b)+0.5, int(b)+0.5
        rect_w, rect_h = int(eff_wd)-1, int(eff_ht)-1

        # Paint the central area offscreen
        cr.push_group()
        for x in xrange(0, eff_wd, step):
            amt = x / eff_wd
            setattr(col, f1, amt)
            setattr(col, f2, 1.0)
            lg = cairo.LinearGradient(b+x, b, b+x, b+eff_ht)
            lg.add_color_stop_rgb(*([0.0] + list(col.get_rgb())))
            setattr(col, f2, 0.0)
            lg.add_color_stop_rgb(*([1.0] + list(col.get_rgb())))
            cr.rectangle(b+x, b, step, eff_ht)
            cr.set_source(lg)
            cr.fill()
        slice_patt = cr.pop_group()

        # Tango-like outline
        cr.set_line_join(cairo.LINE_JOIN_ROUND)
        cr.rectangle(rect_x, rect_y, rect_w, rect_h)
        cr.set_line_width(self.OUTLINE_WIDTH)
        cr.set_source_rgba(*self.OUTLINE_RGBA)
        cr.stroke()

        # The main area
        cr.set_source(slice_patt)
        cr.paint()

        # Tango-like highlight over the top
        cr.rectangle(rect_x, rect_y, rect_w, rect_h)
        cr.set_line_width(self.EDGE_HIGHLIGHT_WIDTH)
        cr.set_source_rgba(*self.EDGE_HIGHLIGHT_RGBA)
        cr.stroke()

    def get_background_validity(self):
        col = HSVColor(color=self.get_managed_color())
        f0 = self.__cube._faces[0]
        return f0, getattr(col, f0)

    def get_color_at_position(self, x, y):
        alloc = self.get_allocation()
        b = self.BORDER_WIDTH
        wd = alloc.width
        ht = alloc.height
        eff_wd = wd - 2*b
        eff_ht = ht - 2*b
        f1_amt = clamp((x-b) / eff_wd, 0, 1)
        f2_amt = clamp((y-b) / eff_ht, 0, 1)
        col = HSVColor(color=self.get_managed_color())
        f1, f2 = self.__get_faces()
        f2_amt = 1.0 - f2_amt
        setattr(col, f1, f1_amt)
        setattr(col, f2, f2_amt)
        return col

    def get_position_for_color(self, col):
        col = HSVColor(color=col)
        f1, f2 = self.__get_faces()
        f1_amt = getattr(col, f1)
        f2_amt = getattr(col, f2)
        f2_amt = 1.0 - f2_amt
        alloc = self.get_allocation()
        b = self.BORDER_WIDTH
        wd = alloc.width
        ht = alloc.height
        eff_wd = wd - 2*b
        eff_ht = ht - 2*b
        x = b + f1_amt*eff_wd
        y = b + f2_amt*eff_ht
        return x, y

    def paint_foreground_cb(self, cr, wd, ht):
        x, y = self.get_position_for_color(self.get_managed_color())
        draw_marker_circle(cr, x, y)


if __name__ == '__main__':
    import os
    import sys
    from adjbases import ColorManager
    mgr = ColorManager(prefs={}, datapath='.')
    cube = HSVSquarePage()
    cube.set_color_manager(mgr)
    mgr.set_color(RGBColor(0.3, 0.6, 0.7))
    if len(sys.argv) > 1:
        icon_name = cube.get_page_icon_name()
        for dir_name in sys.argv[1:]:
            cube.save_icon_tree(dir_name, icon_name)
    else:
        # Interactive test
        window = Gtk.Window()
        window.add(cube.get_page_widget())
        window.set_title(os.path.basename(sys.argv[0]))
        window.connect("destroy", lambda *a: Gtk.main_quit())
        window.show_all()
        Gtk.main()



# This file is part of MyPaint.
# Copyright (C) 2007-2011 by Martin Renold <martinxyz@gmx.ch>
# Copyright (C) 2011-2013 by Andrew Chadwick <a.t.chadwick@gmail.com>
# Copyright (C) 2014-2019 by The MyPaint Team
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

from __future__ import division, print_function

import logging
from math import pi

from lib.gibindings import Gtk, Gdk

from lib.helpers import clamp

logger = logging.getLogger(__name__)

RADIUS = 2


class CurveWidget(Gtk.DrawingArea):
    """Widget for modifying a (restricted) nonlinear curve.
    """
    __gtype_name__ = 'CurveWidget'
    _SNAP_TO = tuple(float(n)/100 for n in range(0, 105, 5))
    _WHINED_ABOUT_ALPHA = False

    def __init__(self, changed_cb=None, magnetic=True, npoints=None,
                 ylockgroups=()):
        Gtk.DrawingArea.__init__(self)
        self.points = [(0.0, 0.2), (.25, .5), (.75, .75), (1.0, 1.0)]
        self._ylock = {}
        self.ylockgroups = ylockgroups

        self.maxpoints = None
        self._npoints = None
        self.npoints = npoints
        self.grabbed = None
        if changed_cb is None:
            self.changed_cb = lambda *a: None
        else:
            self.changed_cb = changed_cb
        self.magnetic = magnetic

        self.connect("draw", self.draw_cb)

        self.connect("button-press-event", self.button_press_cb)
        self.connect("button-release-event", self.button_release_cb)
        self.connect("motion-notify-event", self.motion_notify_cb)
        self.set_events(Gdk.EventMask.EXPOSURE_MASK |
                        Gdk.EventMask.BUTTON_PRESS_MASK |
                        Gdk.EventMask.BUTTON_RELEASE_MASK |
                        Gdk.EventMask.POINTER_MOTION_MASK
                        )
        self.set_size_request(300, 200)

        self.graypoint = None

    @property
    def npoints(self):
        """If the curve contains a fixed number of points, return that number.

        :return: The point count if the count is fixed, otherwise None
        :rtype: int | None
        """
        return self._npoints

    @npoints.setter
    def npoints(self, n):
        """Set the number of points for a fixed curve, or disable fixedness.

        :param n: The number of points, or None to disable
        :type n: int | None
        """
        self._npoints = n
        self.maxpoints = 64 if not n else n

    @property
    def ylockgroups(self):
        """Get a copy of the y-lock groups used by the curve.

        :return: Dictionary of index-> (i1, i2, ...) associations
        :rtype: dict
        """
        return {k: w for k, w in self._ylock.items()}

    @ylockgroups.setter
    def ylockgroups(self, ylockgroups):
        """Set y-lock groups from a list of index tuples.

        Each tuple should contain a up a set of indices that represent curve
        points that will share the same y-value. When the y-value of one point
        in a group is changed, all other points in that same group are changed.

        :param ylockgroups: List of tuples of indices
        :type ylockgroups: [(int, int, ..)]
        """
        self._ylock.clear()
        for group in ylockgroups:
            for idx in group:
                self._ylock[idx] = tuple(i for i in group if i != idx)

    def eventpoint(self, event_x, event_y):
        width, height = self.get_display_area()
        x, y = event_x, event_y
        x -= RADIUS
        y -= RADIUS
        x = x / width
        y = y / height
        return x, y

    def get_display_area(self):
        alloc = self.get_allocation()
        width, height = alloc.width, alloc.height
        width -= 2*RADIUS
        height -= 2*RADIUS
        width = width / 4 * 4
        height = height / 4 * 4
        return width, height

    def set_point(self, index, value):
        y = value[1]
        self.points[index] = value
        if index in self._ylock:
            for locked_to in self._ylock[index]:
                self.points[locked_to] = (self.points[locked_to][0], y)

    def button_press_cb(self, widget, event):
        if not (self.points or event.button == 1):
            return
        x, y = self.eventpoint(event.x, event.y)

        # Note: Squared distance used for comparisons
        def dist_squared(p):
            return abs(x - p[0])**2 + abs(y - p[1])**2
        points = self.points
        dsq, pos = min((dist_squared(p), i) for i, p in enumerate(points))

        # Unless the number of points are fixed, maxed out, or the intent
        # was to move an existing point, insert a new curve point.
        if not (self.npoints or dsq <= 0.003 or len(points) >= self.maxpoints):
            candidates = [i+1 for i, (px, _) in enumerate(points) if px < x]
            insert_pos = candidates and candidates[-1]
            if insert_pos and insert_pos < len(points):
                points.insert(insert_pos, (x, clamp(y, 0.0, 1.0)))
                pos = insert_pos
                self.queue_draw()

        self.grabbed = pos

    def button_release_cb(self, widget, event):
        if not event.button == 1:
            return
        if self.grabbed:
            if self.points[self.grabbed] is None:
                self.points.pop(self.grabbed)
        self.grabbed = None
        # notify user of the widget
        self.changed_cb(self)

    def motion_notify_cb(self, widget, event):
        if self.grabbed is None:
            return
        x, y = self.eventpoint(event.x, event.y)
        i = self.grabbed
        points = self.points
        # XXX this may fail for non contiguous groups.
        if i in self._ylock:
            i_candidate = None
            if x > points[max(self._ylock[i])][0]:
                i_candidate = max((i,) + self._ylock[i])
            elif x < points[min(self._ylock[i])][0]:
                i_candidate = min((i,) + self._ylock[i])
            if (i_candidate is not None and
                    abs(points[i][0] - points[i_candidate][0]) < 0.001):
                i = i_candidate
        out = False  # by default, points cannot be removed
        if i == len(points) - 1:
            # last point stays right
            left_bound = right_bound = 1.0
        elif i == 0:
            # first point stays left
            left_bound = right_bound = 0.0
        else:
            # other points can be dragged out
            left_bound = points[i-1][0]
            right_bound = points[i+1][0]
            margin = 0.02
            inside_x_bounds = left_bound - margin < x < right_bound + margin
            inside_y_bounds = -0.1 <= y <= 1.1
            out = not (self.npoints or (inside_x_bounds and inside_y_bounds))
        if out:
            points[i] = None
        else:
            y = clamp(y, 0.0, 1.0)
            if self.magnetic:
                x_diff = [abs(x - v) for v in self._SNAP_TO]
                y_diff = [abs(y - v) for v in self._SNAP_TO]
                if min(x_diff) < 0.015 and min(y_diff) < 0.015:
                    y = self._SNAP_TO[y_diff.index(min(y_diff))]
                    x = self._SNAP_TO[x_diff.index(min(x_diff))]
            x = clamp(x, left_bound, right_bound)
            self.set_point(i, (x, y))
        self.queue_draw()

    def draw_cb(self, widget, cr):

        def gdk2rgb(c):
            if c.alpha < 1 and not self.__class__._WHINED_ABOUT_ALPHA:
                logger.warning('The GTK3 style is reporting a color with '
                               'alpha less than 1. This should be harmless, '
                               'but please report any glitches with the curve '
                               'widget. Adwaita is thought not to suffer from '
                               'this problem.')
                self.__class__._WHINED_ABOUT_ALPHA = True
            return c.red, c.green, c.blue
        style = widget.get_style_context()
        state = widget.get_state_flags()
        fg = gdk2rgb(style.get_color(state))

        width, height = self.get_display_area()
        if width <= 0 or height <= 0:
            return

        # 1-pixel width, align lines with pixels
        # (but filled rectangles are off by 0.5px now)
        cr.translate(0.5, 0.5)

        # draw feint grid lines
        cr.set_line_width(0.333)
        cr.set_source_rgb(*fg)
        for i in range(11):
            cr.move_to(RADIUS, i*height/10 + RADIUS)
            cr.line_to(width + RADIUS, i*height/10 + RADIUS)
            cr.move_to(i*width/10 + RADIUS, RADIUS)
            cr.line_to(i*width/10 + RADIUS, height + RADIUS)
            cr.stroke()

        # The graypoint is represented by a dashed vertical line spanning the
        # entire graph, with a gap where a circle marks the vertical position.
        if self.graypoint:
            x1, y1 = self.graypoint
            x1 = int(x1*width) + RADIUS
            y1 = int(y1*height) + RADIUS
            cr.set_line_width(0.5)
            cr.set_dash([height/50.0])
            cr.move_to(x1, RADIUS)
            cr.line_to(x1, y1 - 2 * RADIUS)
            cr.move_to(x1, y1 + 2 * RADIUS)
            cr.line_to(x1, height + RADIUS)
            cr.stroke()
            cr.set_dash([])
            cr.arc(x1, y1, 2 * RADIUS, 0, 2*pi)
            cr.stroke()

        # back to regular weight
        cr.set_line_width(1.0)

        # draw points
        prev_x = prev_y = 0
        for p in self.points:
            if p is None:
                continue
            current_x = int(p[0] * width) + RADIUS
            current_y = int(p[1] * height) + RADIUS

            cr.rectangle(
                current_x-RADIUS-0.5, current_y-RADIUS-0.5,
                2*RADIUS+1, 2*RADIUS+1
            )
            cr.fill()

            # If it's the first point, we won't draw any lines yet
            if p is not self.points[0]:
                cr.move_to(prev_x, prev_y)
                cr.line_to(current_x, current_y)
                cr.stroke()

            prev_x, prev_y = current_x, current_y

        return True


def _test(case=1):
    logging.basicConfig()
    win = Gtk.Window()
    curve = CurveWidget()
    if case == 1:
        curve.ylockgroups = [(1, 2), (3, 4)]
        curve.npoints = 6
        curve.points = [
            (0., 0.), (.2, .5), (.4, .75), (.6, .5), (.8, .3), (1., 1.)
        ]
        curve.graypoint = (0.5, 0.0)
    win.add(curve)
    win.set_title("curve test")
    win.connect("destroy", lambda *a: Gtk.main_quit())
    win.show_all()
    Gtk.main()


if __name__ == '__main__':
    _test()



# This file is part of MyPaint.
# -*- coding: utf-8 -*-
# Copyright (C) 2007-2013 by Martin Renold <martinxyz@gmx.ch>
# Copyright (C) 2013-2019 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

import os
import sys
import json
import logging

from lib.gibindings import GLib

import lib.glib
from lib.eotf import DEFAULT_EOTF
from gui.compatconfig import DEFAULT_CONFIG as COMPAT_CONFIG

logger = logging.getLogger(__name__)


def get_json_config(conf_path):
    """Return user settings read from the settings file
    :param conf_path: The path to the directory containing settings.json
    :type conf_path: str
    :returns: Dict with settings, or the empty dict if the settings file
    cannot be found/read or parsed.
    """
    settingspath = os.path.join(conf_path, u'settings.json')
    logger.debug("Reading app settings from %r", settingspath)
    try:
        # Py3: settings.json has always been UTF-8 even in Py2.
        #
        # The Travis-CI json.loads() from Python 3.4.0 needs
        # unicode strings, always. Later and earlier versions,
        # including Py2 do not need that, if bytes are UTF-8.
        with open(settingspath, "rb") as fp:
            return json.loads(fp.read().decode("utf-8"))
    except IOError:
        logger.warning("Failed to load settings file: %s", settingspath)
    except Exception as e:
        logger.warning("settings.json: %s", str(e))
    logger.warning("Failed to load settings: using defaults")
    return {}


def default_configuration():
    """ Return the default settings
    A subset of these settings are platform dependent.
    """
    if sys.platform == 'win32':
        ud_docs = lib.glib.get_user_special_dir(
            GLib.UserDirectory.DIRECTORY_DOCUMENTS,
        )
        scrappre = os.path.join(ud_docs, u'MyPaint', u'scrap')
    else:
        scrappre = u'~/MyPaint/scrap'
    default_config = {
        'saving.scrap_prefix': scrappre,
        'input.device_mode': 'screen',
        'input.global_pressure_mapping': [(0.0, 1.0), (1.0, 0.0)],
        'input.use_barrel_rotation': True,
        'input.barrel_rotation_subtract_ascension': True,
        'input.barrel_rotation_offset': 0.5,
        'view.default_zoom': 1.0,
        'view.real_alpha_checks': True,
        'ui.hide_menubar_in_fullscreen': True,
        'ui.hide_toolbar_in_fullscreen': True,
        'ui.hide_subwindows_in_fullscreen': True,
        'ui.parts': dict(main_toolbar=True, menubar=True),
        'ui.feedback.scale': False,
        'ui.feedback.last_pos': False,
        'ui.toolbar_items': dict(
            toolbar1_file=True,
            toolbar1_scrap=False,
            toolbar1_edit=True,
            toolbar1_blendmodes=False,
            toolbar1_linemodes=True,
            toolbar1_view_modes=True,
            toolbar1_view_manips=False,
            toolbar1_view_resets=True,
        ),
        'ui.toolbar_icon_size': 'large',
        'ui.dark_theme_variant': True,
        'ui.rendered_tile_cache_size': 16384,
        'saving.default_format': 'openraster',
        'brushmanager.selected_brush': None,
        'brushmanager.selected_groups': [],
        'frame.color_rgba': (0.12, 0.12, 0.12, 0.92),
        'misc.context_restores_color': True,

        'document.autosave_backups': True,
        'document.autosave_interval': 10,

        # configurable EOTF.  Set to 1.0 for legacy non-linear behaviour
        'display.colorspace_EOTF': DEFAULT_EOTF,
        'display.colorspace': "srgb",
        # sRGB is a good default even for OS X since v10.6 / Snow
        # Leopard: http://support.apple.com/en-us/HT3712.
        # Version 10.6 was released in September 2009.

        "scratchpad.last_opened_scratchpad": "",

        # Initial main window positions
        "workspace.layout": {
            "position": dict(x=50, y=32, w=-50, h=-100),
            "autohide": True,
        },

        # Linux defaults.
        # Alt is the normal window resizing/moving key these days,
        # so provide a Ctrl-based equivalent for all alt actions.
        'input.button_mapping': {
            # Note that space is treated as a fake Button2
            # It is time to free up the modifiers and Button1
            # '<Shift>Button1': 'StraightMode',
            # '<Control>Button1': 'ColorPickMode',
            # '<Alt>Button1': 'ColorPickMode',
            'Button2': 'PanViewMode',
            '<Shift>Button2': 'RotateViewMode',
            '<Control>Button2': 'ZoomViewMode',
            '<Alt>Button2': 'ZoomViewMode',
            '<Control><Shift>Button2': 'FrameEditMode',
            '<Alt><Shift>Button2': 'FrameEditMode',
            'Button3': 'ShowPopupMenu',
        },
    }
    default_config.update(COMPAT_CONFIG)
    if sys.platform == 'win32':
        # The Linux wacom driver inverts the button numbers of the
        # pen flip button, because middle-click is the more useful
        # action on Linux. However one of the two buttons is often
        # accidentally hit with the thumb while painting. We want
        # to assign panning to this button by default.
        linux_mapping = default_config["input.button_mapping"]
        default_config["input.button_mapping"] = {}
        for bp, actname in linux_mapping.items():
            bp = bp.replace("Button2", "ButtonTMP")
            bp = bp.replace("Button3", "Button2")
            bp = bp.replace("ButtonTMP", "Button3")
            default_config["input.button_mapping"][bp] = actname
    return default_config



# This file is part of MyPaint.
# Copyright (C) 2008-2013 by Martin Renold <martinxyz@gmx.ch>
# Copyright (C) 2013-2018 by the MyPaint Development Team.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or

"""Freehand drawing modes"""

## Imports

from __future__ import division, print_function
import math
import logging
from collections import deque
from gettext import gettext as _

from lib.gibindings import Gtk
from lib.gibindings import Gdk
from lib.gibindings import GLib
import numpy as np

from gui.tileddrawwidget import TiledDrawWidget

from lib.helpers import clamp
import gui.mode
from .drawutils import spline_4p

logger = logging.getLogger(__name__)


## Class defs

class FreehandMode (gui.mode.BrushworkModeMixin,
                    gui.mode.ScrollableModeMixin,
                    gui.mode.InteractionMode):
    """Freehand drawing mode

    To improve application responsiveness, this mode uses an internal
    queue for capturing input data. The raw motion data from the stylus
    is queued; an idle routine then tidies up this data and feeds it
    onward. The presence of an input capture queue means that long
    queued strokes can be terminated by entering a new mode, or by
    pressing Escape.

    This is the default mode in MyPaint.

    """

    ## Class constants & instance defaults

    ACTION_NAME = 'FreehandMode'
    permitted_switch_actions = set()   # Any action is permitted

    _OPTIONS_WIDGET = None

    IS_LIVE_UPDATEABLE = True

    # Motion queue processing (raw data capture)

    # This controls processing of an internal queue of event data such
    # as the x and y coords, pressure and tilt prior to the strokes
    # rendering.

    MOTION_QUEUE_PRIORITY = GLib.PRIORITY_DEFAULT_IDLE

    # The Right Thing To Do generally is to spend as little time as
    # possible directly handling each event received. Disconnecting
    # stroke rendering from event processing buys the user the ability
    # to quit out of a slowly/laggily rendering stroke if desired.

    ## Initialization

    def __init__(self, ignore_modifiers=True, **args):
        # Ignore the additional arg that flip actions feed us
        super(FreehandMode, self).__init__(**args)
        self._cursor_hidden_tdws = set()
        self._cursor_hidden = None

    ## Metadata

    pointer_behavior = gui.mode.Behavior.PAINT_FREEHAND
    scroll_behavior = gui.mode.Behavior.CHANGE_VIEW

    @classmethod
    def get_name(cls):
        return _(u"Freehand Drawing")

    def get_usage(self):
        return _(u"Paint free-form brush strokes")

    ## Per-TDW drawing state

    class _DrawingState (object):
        """Per-canvas drawing state

        Various kinds of queue for raw data capture or interpolation of
        pressure and tilt.
        """

        def __init__(self):
            object.__init__(self)

            self.last_event_had_pressure = False

            # Raw data which was delivered with an identical timestamp
            # to the previous one.  Happens on Windows due to differing
            # clock granularities (at least, under GTK2).
            self._zero_dtime_motions = []

            # Motion Queue

            # Combined, cleaned-up motion data queued ready for
            # interpolation of missing pressures and tilts, then
            # subsequent rendering. Using a queue makes rendering
            # independent of data gathering.
            self.motion_queue = deque()
            self.motion_processing_cbid = None
            self._last_queued_event_time = 0

            # Queued Event Handling

            # Pressure and tilt interpolation for events which
            # don't have pressure or tilt data.
            self.interp = PressureAndTiltInterpolator()

            # Time of the last-processed event
            self.last_handled_event_time = 0

            # Debugging: number of events processed each second,
            # average times.
            self.avgtime = None

            # Button pressed while drawing
            # Not every device sends button presses, but evdev ones
            # do, and this is used as a workaround for an evdev bug:
            # https://github.com/mypaint/mypaint/issues/29
            self.button_down = None
            self.last_good_raw_pressure = 0.0
            self.last_good_raw_xtilt = 0.0
            self.last_good_raw_ytilt = 0.0
            self.last_good_raw_viewzoom = 0.0
            self.last_good_raw_viewrotation = 0.0
            self.last_good_raw_barrel_rotation = 0.0

        def queue_motion(self, event_data):
            """Append one raw motion event to the motion queue

            :param event_data: Extracted data from an event.
            :type event_data: tuple

            Events are tuples of the form ``(time, x, y, pressure,
            xtilt, ytilt, viewzoom, viewrotation, barrel_rotation)``.
            Times are in milliseconds, and are expressed as ints. ``x``
            and ``y`` are ordinary Python floats, and refer to model
            coordinates. The pressure and tilt values have the meaning
            assigned to them by GDK; if ```pressure`` is None, pressure
            and tilt values will be interpolated from surrounding
            defined values.

            Zero-dtime events are detected and cleaned up here.

            """

            (time, x, y, pressure,
             xtilt, ytilt, viewzoom,
             viewrotation, barrel_rotation) = event_data
            if time < self._last_queued_event_time:
                logger.warning('Time is running backwards! Corrected.')
                time = self._last_queued_event_time

            if time == self._last_queued_event_time:
                # On Windows, GTK timestamps have a resolution around
                # 15ms, but tablet events arrive every 8ms.
                # https://gna.org/bugs/index.php?16569
                zdata = (x, y, pressure, xtilt, ytilt, viewzoom,
                         viewrotation, barrel_rotation)
                self._zero_dtime_motions.append(zdata)
            else:
                # Queue any previous events that had identical
                # timestamps, linearly interpolating their times.
                if self._zero_dtime_motions:
                    dtime = time - self._last_queued_event_time
                    if dtime > 100:
                        # Really old events; don't associate them with
                        # the new one.
                        zt = time - 100.0
                        interval = 100.0
                    else:
                        zt = self._last_queued_event_time
                        interval = dtime
                    step = interval / (len(self._zero_dtime_motions) + 1)

                    for (zx, zy, zp, zxt,
                         zyt, zvz, zvr, zbr) in self._zero_dtime_motions:
                        zt += step
                        zevent_data = (zt, zx, zy, zp, zxt, zyt, zvz, zvr, zbr)
                        self.motion_queue.append(zevent_data)
                    # Reset the backlog buffer
                    self._zero_dtime_motions = []
                # Queue this event too
                self.motion_queue.append(event_data)
                # Update the timestamp used above
                self._last_queued_event_time = time

        def next_processing_events(self):
            """Fetches zero or more events to process from the queue"""
            if len(self.motion_queue) > 0:
                event = self.motion_queue.popleft()
                for ievent in self.interp.feed(*event):
                    yield ievent

    def _reset_drawing_state(self):
        """Resets all per-TDW drawing state"""
        self._drawing_state = {}

    def _get_drawing_state(self, tdw):
        drawstate = self._drawing_state.get(tdw, None)
        if drawstate is None:
            drawstate = self._DrawingState()
            self._drawing_state[tdw] = drawstate
        return drawstate

    ## Mode stack & current mode

    def enter(self, doc, **kwds):
        """Enter freehand mode"""
        super(FreehandMode, self).enter(doc, **kwds)
        self._drawing_state = {}
        self._reset_drawing_state()
        self._debug = (logger.getEffectiveLevel() == logging.DEBUG)

    def leave(self, **kwds):
        """Leave freehand mode"""
        self._reset_drawing_state()
        self._reinstate_drawing_cursor(tdw=None)
        super(FreehandMode, self).leave(**kwds)

    ## Special cursor state while there's pressure

    def _hide_drawing_cursor(self, tdw):
        """Hide the cursor while painting, if configured to.

        :param tdw: Canvas widget to hide the cursor on.
        :type tdw: TiledDrawWidget

        """
        if tdw in self._cursor_hidden_tdws:
            return
        if not tdw.app:
            return
        if not tdw.app.preferences.get("ui.hide_cursor_while_painting"):
            return

        if self._cursor_hidden is None:
            window = tdw.get_window()
            cursor = Gdk.Cursor.new_for_display(
                window.get_display(), Gdk.CursorType.BLANK_CURSOR)
            self._cursor_hidden = cursor

        tdw.set_override_cursor(self._cursor_hidden)
        self._cursor_hidden_tdws.add(tdw)

    def _reinstate_drawing_cursor(self, tdw=None):
        """Un-hide any hidden cursors.

        :param tdw: Canvas widget to reset. None means all affected.
        :type tdw: TiledDrawWidget

        """
        if tdw is None:
            for tdw in self._cursor_hidden_tdws:
                tdw.set_override_cursor(None)
            self._cursor_hidden_tdws.clear()
        elif tdw in self._cursor_hidden_tdws:
            tdw.set_override_cursor(None)
            self._cursor_hidden_tdws.remove(tdw)

    ## Input handlers

    def button_press_cb(self, tdw, event):
        result = False
        current_layer = tdw.doc.layer_stack.current
        if (current_layer.get_paintable() and event.button == 1
                and event.type == Gdk.EventType.BUTTON_PRESS):
            # Single button press
            # Stroke started, notify observers
            self.doc.input_stroke_started(event)
            # Mouse button pressed (while painting without pressure
            # information)
            drawstate = self._get_drawing_state(tdw)
            if not drawstate.last_event_had_pressure:
                # For the mouse we don't get a motion event for
                # "pressure" changes, so we simulate it. (Note: we can't
                # use the event's button state because it carries the
                # old state.)
                self.motion_notify_cb(tdw, event,
                                      fakepressure=tdw.app.fakepressure)

            drawstate.button_down = event.button
            drawstate.last_good_raw_pressure = 0.0
            drawstate.last_good_raw_xtilt = 0.0
            drawstate.last_good_raw_ytilt = 0.0
            drawstate.last_good_raw_viewzoom = 0.0
            drawstate.last_good_raw_viewrotation = 0.0
            drawstate.last_good_raw_barrel_rotation = 0.0

            # Hide the cursor if configured to
            self._hide_drawing_cursor(tdw)

            result = True
        return (super(FreehandMode, self).button_press_cb(tdw, event)
                or result)

    def button_release_cb(self, tdw, event):
        result = False
        current_layer = tdw.doc.layer_stack.current
        if current_layer.get_paintable() and event.button == 1:
            # See comment above in button_press_cb.
            drawstate = self._get_drawing_state(tdw)
            if not drawstate.last_event_had_pressure:
                self.motion_notify_cb(tdw, event, fakepressure=0.0)
            # Notify observers after processing the event
            self.doc.input_stroke_ended(event)

            drawstate.button_down = None
            drawstate.last_good_raw_pressure = 0.0
            drawstate.last_good_raw_xtilt = 0.0
            drawstate.last_good_raw_ytilt = 0.0

            # Reinstate the normal cursor if it was hidden
            self._reinstate_drawing_cursor(tdw)

            result = True
        return (super(FreehandMode, self).button_release_cb(tdw, event)
                or result)

    def motion_notify_cb(self, tdw, event, fakepressure=None):
        """Motion event handler: queues raw input and returns

        :param tdw: The TiledDrawWidget receiving the event
        :param event: the MotionNotify event being handled
        :param fakepressure: fake pressure to use if no real pressure

        Fake pressure is passed with faked motion events, e.g.
        button-press and button-release handlers for mouse events.

        """

        # Do nothing if painting is inactivated
        current_layer = tdw.doc._layers.current
        if not (tdw.is_sensitive and current_layer.get_paintable()):
            return False

        # If the device has changed and the last pressure value from the
        # previous device is not equal to 0.0, this can leave a visible
        # stroke on the layer even if the 'new' device is not pressed on
        # the tablet and has a pressure axis == 0.0.  Resetting the brush
        # when the device changes fixes this issue, but there may be a
        # much more elegant solution that only resets the brush on this
        # edge-case.
        same_device = True
        if tdw.app is not None:
            device = event.get_source_device()
            same_device = tdw.app.device_monitor.device_used(device)
            if not same_device:
                tdw.doc.brush.reset()

        # Extract the raw readings for this event
        x = event.x
        y = event.y
        time = event.time
        pressure = event.get_axis(Gdk.AxisUse.PRESSURE)
        xtilt = event.get_axis(Gdk.AxisUse.XTILT)
        ytilt = event.get_axis(Gdk.AxisUse.YTILT)
        viewzoom = tdw.scale
        viewrotation = tdw.rotation
        barrel_rotation = event.get_axis(Gdk.AxisUse.WHEEL)
        state = event.state

        # Workaround for buggy evdev behaviour.
        # Events sometimes get a zero raw pressure reading when the
        # pressure reading has not changed. This results in broken
        # lines. As a workaround, forbid zero pressures if there is a
        # button pressed down, and substitute the last-known good value.
        # Detail: https://github.com/mypaint/mypaint/issues/29
        drawstate = self._get_drawing_state(tdw)
        if drawstate.button_down is not None:
            if pressure == 0.0:
                pressure = drawstate.last_good_raw_pressure
            elif pressure is not None and np.isfinite(pressure):
                drawstate.last_good_raw_pressure = pressure

        # Ensure each event has a defined pressure
        if pressure is not None:
            # Using the reported pressure. Apply some sanity checks
            if not np.isfinite(pressure):
                # infinity/nan: use button state (instead of clamping in
                # brush.hpp) https://gna.org/bugs/?14709
                pressure = None
            else:
                pressure = clamp(pressure, 0.0, 1.0)
            drawstate.last_event_had_pressure = True

        # Fake the pressure if we have none, or if infinity was reported
        if pressure is None:
            if fakepressure is not None:
                pressure = clamp(fakepressure, 0.0, 1.0)
            else:
                pressure = (
                    (state & Gdk.ModifierType.BUTTON1_MASK) and
                    tdw.app.fakepressure or 0.0)
            drawstate.last_event_had_pressure = False

        # Check whether tilt is present.  For some tablets without
        # tilt support GTK reports a tilt axis with value nan, instead
        # of None.  https://gna.org/bugs/?17084
        if xtilt is None or ytilt is None or not np.isfinite(xtilt + ytilt):
            xtilt = 0.0
            ytilt = 0.0

        # Switching from a non-tilt device to a device which reports
        # tilt can cause GDK to return out-of-range tilt values, on X11.
        xtilt = clamp(xtilt, -1.0, 1.0)
        ytilt = clamp(ytilt, -1.0, 1.0)

        tilt_ascension = 0.5 * math.atan2(-xtilt, ytilt) / math.pi

        # Offset barrel rotation if wanted
        # This could be used to correct for different devices,
        # Left vs Right handed, etc.
        b_offset = tdw.app.preferences.get("input.barrel_rotation_offset")
        if (barrel_rotation is not None):
            barrel_rotation = (barrel_rotation + b_offset) % 1.0

        # barrel_rotation is likely affected by ascension (a bug?)
        # lets compensate but allow disabling
        if (barrel_rotation is not None and tdw.app.preferences.get(
           "input.barrel_rotation_subtract_ascension")):
            barrel_rotation = (barrel_rotation - tilt_ascension) % 1.0

        # If WHEEL is missing (barrel_rotation)
        # Use the fakerotation controller to allow keyboard control
        # We can't trust None so also look at preference
        if (barrel_rotation is None
           or not tdw.app.preferences.get("input.use_barrel_rotation")):
            barrel_rotation = (tdw.app.fakerotation + b_offset) % 1.0

        # Evdev workaround. X and Y tilts suffer from the same
        # problem as pressure for fancier devices.
        if drawstate.button_down is not None:
            if xtilt == 0.0:
                xtilt = drawstate.last_good_raw_xtilt
            else:
                drawstate.last_good_raw_xtilt = xtilt
            if ytilt == 0.0:
                ytilt = drawstate.last_good_raw_ytilt
            else:
                drawstate.last_good_raw_ytilt = ytilt

        if tdw.mirrored:
            xtilt *= -1.0
            barrel_rotation *= -1.0

        # Apply pressure mapping if we're running as part of a full
        # MyPaint application (and if there's one defined).
        if tdw.app is not None and tdw.app.pressure_mapping:
            pressure = tdw.app.pressure_mapping(pressure)

        # Apply any configured while-drawing cursor
        if pressure > 0:
            self._hide_drawing_cursor(tdw)
        else:
            self._reinstate_drawing_cursor(tdw)

        # Queue this event
        x, y = tdw.display_to_model(x, y)

        event_data = (time, x, y, pressure,
                      xtilt, ytilt, viewzoom,
                      viewrotation, barrel_rotation)
        drawstate.queue_motion(event_data)
        # Start the motion event processor, if it isn't already running
        if not drawstate.motion_processing_cbid:
            cbid = GLib.idle_add(
                self._motion_queue_idle_cb,
                tdw,
                priority = self.MOTION_QUEUE_PRIORITY,
            )
            drawstate.motion_processing_cbid = cbid

    ## Motion queue processing

    def _motion_queue_idle_cb(self, tdw):
        """Idle callback; processes each queued event"""
        drawstate = self._get_drawing_state(tdw)
        # Stop if asked to stop
        if drawstate.motion_processing_cbid is None:
            drawstate.motion_queue = deque()
            return False
        # Forward one or more motion events to the canvas
        for event in drawstate.next_processing_events():
            self._process_queued_event(tdw, event)
        # Stop if the queue is now empty
        if len(drawstate.motion_queue) == 0:
            drawstate.motion_processing_cbid = None
            return False
        # Otherwise, continue being invoked
        return True

    def _process_queued_event(self, tdw, event_data):
        """Process one motion event from the motion queue"""
        drawstate = self._get_drawing_state(tdw)
        (time, x, y, pressure, xtilt, ytilt, viewzoom,
         viewrotation, barrel_rotation) = event_data
        model = tdw.doc

        # Calculate time delta for the brush engine
        last_event_time = drawstate.last_handled_event_time
        drawstate.last_handled_event_time = time
        if not last_event_time:
            return
        dtime = (time - last_event_time) / 1000.0
        if self._debug:
            cavg = drawstate.avgtime
            if cavg is not None:
                tavg, nevents = cavg
                nevents += 1
                tavg += (dtime - tavg) / nevents
            else:
                tavg = dtime
                nevents = 1
            if ((nevents * tavg) > 1.0) and nevents > 20:
                logger.debug("Processing at %d events/s (t_avg=%0.3fs)",
                             nevents, tavg)
                drawstate.avgtime = None
            else:
                drawstate.avgtime = (tavg, nevents)

        current_layer = model._layers.current
        if not current_layer.get_paintable():
            return

        # Feed data to the brush engine.  Pressure and tilt cleanup
        # needs to be done here to catch all forwarded data after the
        # earlier interpolations. The interpolation method used for
        # filling in missing axis data is known to generate
        # OverflowErrors for legitimate but pathological input streams.
        # https://github.com/mypaint/mypaint/issues/344

        pressure = clamp(pressure, 0.0, 1.0)
        xtilt = clamp(xtilt, -1.0, 1.0)
        ytilt = clamp(ytilt, -1.0, 1.0)
        self.stroke_to(model, dtime, x, y, pressure,
                       xtilt, ytilt, viewzoom,
                       viewrotation, barrel_rotation)

        # Update the TDW's idea of where we last painted
        # FIXME: this should live in the model, not the view
        if pressure:
            tdw.set_last_painting_pos((x, y))

    ## Mode options

    def get_options_widget(self):
        """Get the (class singleton) options widget"""
        cls = self.__class__
        if cls._OPTIONS_WIDGET is None:
            widget = FreehandOptionsWidget()
            cls._OPTIONS_WIDGET = widget
        return cls._OPTIONS_WIDGET


class FreehandOptionsWidget (gui.mode.PaintingModeOptionsWidgetBase):
    """Configuration widget for freehand mode"""

    def init_specialized_widgets(self, row):
        cname = "slow_tracking"
        label = Gtk.Label()
        # TRANSLATORS: Short alias for "Slow position tracking". This is
        # TRANSLATORS: used on the options panel.
        label.set_text(_("Smooth:"))
        label.set_alignment(1.0, 0.5)
        label.set_hexpand(False)
        self.adjustable_settings.add(cname)
        adj = self.app.brush_adjustment[cname]
        scale = Gtk.Scale.new(Gtk.Orientation.HORIZONTAL, adj)
        scale.set_draw_value(False)
        scale.set_hexpand(True)
        self.attach(label, 0, row, 1, 1)
        self.attach(scale, 1, row, 1, 1)
        row += 1
        cname = "fakepressure"
        label = Gtk.Label()
        # TRANSLATORS: Short alias for "Fake Pressure (for mouse)". This is
        # TRANSLATORS: used on the options panel.
        label.set_text(_("Pressure:"))
        label.set_alignment(1.0, 0.5)
        label.set_hexpand(False)
        changed_cb = self._fakepressure_value_changed_cb
        adj = Gtk.Adjustment(value=0.5, lower=0.0, upper=1.0,
                             step_increment=0.01, page_increment=0.1)
        self.app.fake_adjustment['fakepressure'] = adj
        adj.connect("value-changed", changed_cb)
        scale = Gtk.Scale.new(Gtk.Orientation.HORIZONTAL, adj)
        scale.set_draw_value(False)
        scale.set_hexpand(True)
        self.attach(label, 0, row, 1, 1)
        self.attach(scale, 1, row, 1, 1)
        row += 1
        cname = "fakerotation"
        label = Gtk.Label()
        # TRANSLATORS: Short alias for "Fake Barrel Rotation (for mouse)".
        # TRANSLATORS: This is used on the options panel.
        label.set_text(_("Twist:"))
        label.set_alignment(1.0, 0.5)
        label.set_hexpand(False)
        changed_cb = self._fakerotation_value_changed_cb
        adj = Gtk.Adjustment(value=0.5, lower=0.0, upper=1.0,
                             step_increment=0.0625, page_increment=0.25)
        self.app.fake_adjustment['fakerotation'] = adj
        adj.connect("value-changed", changed_cb)
        scale = Gtk.Scale.new(Gtk.Orientation.HORIZONTAL, adj)
        scale.set_draw_value(False)
        scale.set_hexpand(True)
        self.attach(label, 0, row, 1, 1)
        self.attach(scale, 1, row, 1, 1)
        row += 1
        return row

    def _fakepressure_value_changed_cb(self, adj):
        """Updates fakepressure when the user tweaks it using a scale"""
        newvalue = adj.get_value()
        self.app.fakepressure = newvalue

    def fakepressure_modified_cb(self, value):
        """Updates the fakepressure slider when changed elsewhere"""
        adj = self.app.fake_adjustment.get('fakepressure', None)
        if adj is not None:
            adj.set_value(value)

    def _fakerotation_value_changed_cb(self, adj):
        """Updates fakerotation when the user tweaks it using a scale"""
        newvalue = adj.get_value()
        self.app.fakerotation = newvalue

    def fakerotation_modified_cb(self, value):
        """Updates the fakerotation slider when changed elsewhere"""
        adj = self.app.fake_adjustment.get('fakerotation', None)
        if adj is not None:
            adj.set_value(value)


class PressureAndTiltInterpolator (object):
    """Interpolates event sequences, filling in null pressure/tilt data

    The interpolator operates almost as a filter. Feed the interpolator
    an extra zero-pressure event at button-release time to generate a
    nice tailoff for mouse users. The interpolator is sensitive to
    transitions between nonzero and zero effective pressure in both
    directions. These transitions clear out just enough history to avoid
    hook-off and lead-in artifacts.

    >>> interp = PressureAndTiltInterpolator()
    >>> raw_data = interp._TEST_DATA
    >>> all([len(t) == 9 for t in raw_data])
    True
    >>> any([t for t in raw_data if None in t[3:]])
    True
    >>> cooked_data = []
    >>> for raw_event in raw_data:
    ...    for cooked_event in interp.feed(*raw_event):
    ...        cooked_data.append(cooked_event)
    >>> any([t for t in cooked_data if None in t[3:]])
    False
    >>> len(cooked_data) <= len(raw_data)
    True
    >>> all([(t in cooked_data) for t in raw_data
    ...      if None not in t[3:]])
    True
    >>> any([t for t in cooked_data if 70 < t[0] < 110])
    False
    >>> len([t for t in cooked_data if t[0] in (70, 110)]) == 2
    True

    """

    # Test data:

    _TEST_DATA = [
        # These 2 events will be dropped (no prior state with pressure).
        (3, 0.3, 0.3, None, None, None, None, None, None),
        (7, 0.7, 0.7, None, None, None, None, None, None),
        (10, 1.0, 1.0, 0.33, 0.0, 0.5, 1.0, 0.0, 0.0),
        # Gaps between defined data like this one will have those
        # None entries filled in.
        (13, 1.3, 1.3, None, None, None, None, None, None),
        (15, 1.5, 1.5, None, None, None, None, None, None),
        (17, 1.7, 1.7, None, None, None, None, None, None),
        (20, 2.0, 2.0, 0.45, 0.1, 0.4, 1.0, 0.0, 0.0),
        (23, 2.3, 2.3, None, None, None, None, None, None),
        (27, 2.7, 2.7, None, None, None, None, None, None),
        (30, 3.0, 3.0, 0.50, 0.2, 0.3, 1.0, 0.0, 0.0),
        (33, 3.3, 3.3, None, None, None, None, None, None),
        (37, 3.7, 3.7, None, None, None, None, None, None),
        (40, 4.0, 4.0, 0.40, 0.3, 0.2, 1.0, 0.0, 0.0),
        (44, 4.4, 4.4, None, None, None, None, None, None),
        (47, 4.7, 4.7, None, None, None, None, None, None),
        (50, 5.0, 5.0, 0.30, 0.5, 0.1, 1.0, 0.0, 0.0),
        (53, 5.3, 5.3, None, None, None, None, None, None),
        (57, 5.7, 5.7, None, None, None, None, None, None),
        (60, 6.0, 6.0, 0.11, 0.4, 0.0, 1.0, 0.0, 0.0),
        (63, 6.3, 6.3, None, None, None, None, None, None),
        (67, 6.7, 6.7, None, None, None, None, None, None),
        # Down to zero pressure...
        (70, 7.0, 7.0, 0.00, 0.2, 0.0, 1.0, 0.0, 0.0),
        # .. followed by a null-pressure sequence.
        # That means that this gap will be skipped over till an
        # event with a defined pressure comes along.
        (73, 7.0, 7.0, None, None, None, None, None, None),
        (78, 50.0, 50.0, None, None, None, None, None, None),
        (83, 110.0, 110.0, None, None, None, None, None, None),
        (88, 120.0, 120.0, None, None, None, None, None, None),
        (93, 130.0, 130.0, None, None, None, None, None, None),
        (98, 140.0, 140.0, None, None, None, None, None, None),
        (103, 150.0, 150.0, None, None, None, None, None, None),
        (108, 160.0, 160.0, None, None, None, None, None, None),
        # Normally, event tuples won't be altered or have extra events
        # inserted between them.
        (110, 170.0, 170.0, 0.11, 0.1, 0.0, 1.0, 0.0, 0.0),
        (120, 171.0, 171.0, 0.33, 0.0, 0.0, 1.0, 0.0, 0.0),
        (130, 172.0, 172.0, 0.00, 0.0, 0.0, 1.0, 0.0, 0.0)
    ]

    # Construction:

    def __init__(self):
        """Instantiate with a clear internal state"""
        object.__init__(self)
        # Events with all axis data present, forming control points
        self._pt0_prev = None
        self._pt0 = None
        self._pt1 = None
        self._pt1_next = None
        # Null-axis event sequences
        self._np = []
        self._np_next = []

    # Internals:

    def _clear(self):
        """Reset to the initial clean state"""
        self._pt0_prev = None
        self._pt0 = None
        self._pt1 = None
        self._pt1_next = None
        self._np = []
        self._np_next = []

    def _step(self):
        """Step the interpolation parameters forward"""
        self._pt0_prev = self._pt0
        self._pt0 = self._pt1
        self._pt1 = self._pt1_next
        self._pt1_next = None
        self._np = self._np_next
        self._np_next = []

    def _interpolate_p0_p1(self):
        """Interpolate between p0 and p1, but do not step or clear"""
        pt0p, pt0 = self._pt0_prev, self._pt0
        pt1, pt1n = self._pt1, self._pt1_next
        can_interp = (pt0 is not None and pt1 is not None and
                      len(self._np) > 0)
        if can_interp:
            if pt0p is None:
                pt0p = pt0
            if pt1n is None:
                pt1n = pt1
            t0 = pt0[0]
            t1 = pt1[0]
            dt = t1 - t0
            can_interp = dt > 0
        if can_interp:
            for event in self._np:
                t, x, y = event[0:3]
                p, xt, yt, vz, vr, br = spline_4p(
                    (t - t0) / dt,
                    np.array(pt0p[3:]), np.array(pt0[3:]),
                    np.array(pt1[3:]), np.array(pt1n[3:])
                )
                yield (t, x, y, p, xt, yt, vz, vr, br)
        if pt1 is not None:
            yield pt1

    def _interpolate_and_step(self):
        """Internal: interpolate & step forward or clear"""
        for ievent in self._interpolate_p0_p1():
            yield ievent
        if ((self._pt1_next[3] > 0.0) and
                (self._pt1 is not None) and
                (self._pt1[3] <= 0.0)):
            # Transitions from zero to nonzero pressure
            # Clear history to avoid artifacts
            self._pt0_prev = None   # ignore the current pt0
            self._pt0 = self._pt1
            self._pt1 = self._pt1_next
            self._pt1_next = None
            self._np = []           # drop the buffer we've built up too
            self._np_next = []
        elif ((self._pt1_next[3] <= 0.0) and
              (self._pt1 is not None) and (self._pt1[3] > 0.0)):
            # Transitions from nonzero to zero pressure
            # Tail off neatly by doubling the zero-pressure event
            self._step()
            self._pt1_next = self._pt1
            for ievent in self._interpolate_p0_p1():
                yield ievent
            # Then clear history
            self._clear()
        else:
            # Normal forward of control points and event buffers
            self._step()

    # Public methods:

    def feed(self, time, x, y, pressure, xtilt, ytilt, viewzoom,
             viewrotation, barrel_rotation):
        """Feed in an event, yielding zero or more interpolated events

        :param time: event timestamp, integer number of milliseconds
        :param x: Horizontal coordinate of the event, in model space
        :type x: float
        :param y: Vertical coordinate of the event, in model space
        :type y: float
        :param pressure: Effective pen pressure, [0.0, 1.0]
        :param xtilt: Pen tilt in the model X direction, [-1.0, 1.0]
        :param ytilt: Pen tilt in the model's Y direction, [-1.0, 1.0]
        :param viewzoom: The view's current zoom level, [0, 64]
        :param viewrotation: The view's current rotation, [-180.0, 180.0]
        :param barrel_rotation: The stylus barrel rotation, [0.0, 1.0]
        :returns: Iterator of event tuples

        Event tuples have the form (TIME, X, Y, PRESSURE, XTILT, YTILT,
        VIEWZOOM, VIEWROTATION, BARREL_ROTATION).
        """
        if None in (pressure, xtilt, ytilt, viewzoom,
                    viewrotation, barrel_rotation):
            self._np_next.append((time, x, y, pressure, xtilt, ytilt, viewzoom,
                                  viewrotation, barrel_rotation))
        else:
            self._pt1_next = (time, x, y, pressure, xtilt, ytilt, viewzoom,
                              viewrotation, barrel_rotation)
            for t, x, y, p, xt, yt, vz, vr, br in self._interpolate_and_step():
                yield (t, x, y, p, xt, yt, vz, vr, br)


## Module tests

def _test():
    import doctest
    doctest.testmod()
    interp = PressureAndTiltInterpolator()
    # Emit CSV for ad-hoc plotting
    print("time,x,y,pressure,xtilt,ytilt,viewzoom,viewrotation,barrel_rotation")
    for event in interp._TEST_DATA:
        for data in interp.feed(*event):
            print(",".join([str(c) for c in data]))


if __name__ == '__main__':
    _test()
